from typing import Dict
from typing import List
from typing import cast
from typing import NewType

from logging import Logger
from logging import getLogger

from datetime import datetime

from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutField import PyutField
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutParameter import PyutParameter
from pyutmodel.PyutType import PyutType

from pyutmodel.PyutVisibilityEnum import PyutVisibilityEnum

MethodsCodeType = NewType('MethodsCodeType', Dict[str, List[str]])


class PyutToPython:

    MAX_WIDTH:            int = 120
    CLASS_COMMENTS_START: str = '"""'
    CLASS_COMMENTS_END:   str = '"""'
    SINGLE_TAB:           str = '    '

    SPECIAL_PYTHON_CONSTRUCTOR: str = '__init__'

    """
    Reads the Pyut data model in order to generated syntactically correct Python code
    """
    def __init__(self):

        self.logger: Logger = getLogger(__name__)

    def generateTopCode(self) -> List[str]:

        now = datetime.now()

        dateTimeStr: str = now.strftime("%Y-%m-%d %H:%M:%S")
        topCode: List[str] = [
            f'# \n',
            f'# Generated by PyutToPython 1.0.0\n',
            f'# -- La Vida Buena, LLC\n',
            f'# {dateTimeStr}\n',
            f'# \n'
            ]

        return topCode

    def generateClassStanza(self, pyutClass: PyutClass) -> str:
        """
        Generates something like this

        ```python
            class Car:
        ```
        or with inheritance

        ```python
            class ElectricCar(BaseVehicle, Car):
        ```
`
        Args:
            pyutClass:   The data model class

        Returns:
            The Python class start stanza
        """
        generatedCode:     str             = f'class {pyutClass.name}'
        parentPyutClasses: List[PyutClass] = cast(List[PyutClass], pyutClass.getParents())

        if len(parentPyutClasses) > 0:  # Add parents
            generatedCode = f'{generatedCode}('
            for i in range(len(parentPyutClasses)):
                generatedCode = f'{generatedCode}{parentPyutClasses[i].name}'
                if i < len(parentPyutClasses) - 1:
                    generatedCode = f'{generatedCode},'
            generatedCode = f'{generatedCode})'
        generatedCode = f'{generatedCode}:\n'
        generatedCode = f'{generatedCode}{self.__indentStr(PyutToPython.CLASS_COMMENTS_START)}\n'
        generatedCode = f'{generatedCode}{self.__indentStr(pyutClass.description)}\n'   # TODO need to split lines according to MAX_WIDTH
        generatedCode = f'{generatedCode}{self.__indentStr(PyutToPython.CLASS_COMMENTS_END)}\n'

        return generatedCode

    def generateMethodsCode(self, pyutClass: PyutClass) -> MethodsCodeType:
        """
        Return a dictionary of method code for a given class

        Args:
            pyutClass:  The data model class for which we have to generate a bunch of code

        Returns:
            A bunch of code that is the code for this class; The map key is the method name the
            value is a list of the method code
        """
        clsMethods: MethodsCodeType = MethodsCodeType({})
        for pyutMethod in pyutClass.methods:
            # Separation
            txt = ""
            lstCodeMethod = [txt]

            # Get code
            subCode:       List[str] = self.generateASingleMethodsCode(pyutMethod)
            lstCodeMethod += self.indent(subCode)

            clsMethods[pyutMethod.name] = lstCodeMethod

        # Create method __init__ if it does not exist
        if PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR not in clsMethods:
            clsMethods = self.generatePythonInitMethod(clsMethods)

        if len(pyutClass.fields) > 0:
            # Add fields

            clsInitMethod: List[str] = clsMethods[PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR]
            for pyutField in pyutClass.fields:
                clsInitMethod.append(self.__indentStr(self.__indentStr(self.generateFieldPythonCode(pyutField))))
            # This is the actual entry;  So no need to put it back in clsMethods
            clsInitMethod.append('\n')

        return clsMethods

    def generatePythonInitMethod(self, clsMethods: MethodsCodeType) -> MethodsCodeType:
        """

        Args:
            clsMethods:  The current in progress dictionary

        Returns:  The updated dictionary
        """
        lstCodeMethod = []
        subCode = self.generateASingleMethodsCode(PyutMethod(PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR), False)

        for el in self.indent(subCode):
            lstCodeMethod.append(str(el))

        clsMethods[PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR] = lstCodeMethod

        return clsMethods

    def generateASingleMethodsCode(self, pyutMethod: PyutMethod, writePass: bool = True) -> List[str]:
        """
        Generate the Python code for the input method

        Args:
            pyutMethod:    The PyutMethod for which we will generate code
            writePass:  If `True` write `pass` in the code

        Returns:
            A list that is the generated code
        """
        methodCode:  List[str] = []

        currentCode: str = self._generateMethodDefinitionStanza(pyutMethod)
        # Add parameters (parameter, parameter, parameter, ...)
        params = pyutMethod.parameters
        currentCode = self._generateParametersCode(currentCode, params)
        currentCode = f'{currentCode})'

        returnType: PyutType = pyutMethod.returnType
        if returnType is not None and returnType.value != '':
            currentCode = f'{currentCode} -> {returnType.value}'

        currentCode = f'{currentCode}:\n'

        # Add to the method code
        methodCode.append(currentCode)

        # Add comments
        methodCode.append(self.__indentStr('"""\n'))
        methodCode = self._generateMethodComments(methodCode, pyutMethod)
        methodCode.append(self.__indentStr('"""\n'))

        if writePass:
            methodCode.append(self.__indentStr('pass\n'))

        methodCode.append('\n')
        return methodCode

    def generateFieldPythonCode(self, pyutField: PyutField):
        """
        Generate the Python code for a given field

        Args:
            pyutField:   The PyutField that is the source of our code generation

        Returns:
            Python Code !!
        """
        fieldCode: str = "self."

        fieldCode = f'{fieldCode}{self.generateVisibilityPrefix(pyutField.visibility)}'
        if pyutField.type.value == '':
            fieldCode = f'{fieldCode}{pyutField.name}'
        else:
            fieldCode = f'{fieldCode}{pyutField.name}: {pyutField.type}'

        value = pyutField.defaultValue
        if value == '':
            fieldCode = f'{fieldCode} = None'
        else:
            fieldCode = f'{fieldCode} = {value}'

        fieldCode = f'{fieldCode}\n'
        return fieldCode

    def generateVisibilityPrefix(self, visibility: PyutVisibilityEnum) -> str:
        """
        Return the python code for the given enumeration value

        Args:
            visibility:

        Returns:
            The Python code that by convention depicts `method` or `field` visibility
        """
        code: str = ''
        if visibility == PyutVisibilityEnum.PUBLIC:
            code = ''
        elif visibility == PyutVisibilityEnum.PROTECTED:
            code = '_'
        elif visibility == PyutVisibilityEnum.PRIVATE:
            code = '__'
        else:
            self.logger.error(f"PyutToPython: Field code not supported : {visibility}")
        self.logger.debug(f"Python code: {code}, for {visibility}")
        return code

    def _generateMethodDefinitionStanza(self, pyutMethod: PyutMethod):
        """
        Follow Python conventions for method visibility

        Something like:
        ```python
            def publicMethod(self)
            def _protectedMethod(self)
            def __privateMethod(self)
        ```
        Args:
            pyutMethod: The method whose code we are generating

        Returns:
            The method start stanza
        """
        currentCode: str = "def "
        if self.__isDunderMethod(pyutMethod.name) is False:
            currentCode = f'{currentCode}{self.generateVisibilityPrefix(pyutMethod.getVisibility())}'
        currentCode = f'{currentCode}{pyutMethod.name}(self'

        return currentCode

    def _generateParametersCode(self, currentCode: str, params: List[PyutParameter]):

        if len(params) > 0:
            currentCode = f'{currentCode}, '
        # Add parameter code
        for i in range(len(params)):
            pyutParam: PyutParameter = params[i]
            numParams: int = len(params)
            paramCode: str = self.__generateParameter(currentParamNumber=i, numberOfParameters=numParams, pyutParam=pyutParam)

            currentCode = self.__addParamToMethodSignature(currentCode, paramCode)
        return currentCode

    def __generateParameter(self, currentParamNumber: int, numberOfParameters: int, pyutParam: PyutParameter) -> str:
        """

        Args:
            currentParamNumber: The current parameter #
            numberOfParameters: The number of parameters the method has
            pyutParam:          What we are generating code from

        Returns:
            Python code for a single parameter
        """

        paramCode: str = ""

        paramCode = f'{paramCode}{pyutParam.name}'

        paramType: PyutType = pyutParam.type
        if paramType is not None and paramType.value != '':
            paramCode = f'{paramCode}: {paramType.value}'
        if pyutParam.defaultValue is not None:
            paramCode = f'{paramCode} = {pyutParam.defaultValue}'
        if currentParamNumber < numberOfParameters - 1:
            paramCode = f'{paramCode}, '

        return paramCode

    def _generateMethodComments(self, methodCode, pyutMethod):

        methodCode.append(self.__indentStr('(TODO : add description)\n\n'))

        params: List[PyutParameter] = pyutMethod.parameters

        if len(params) > 0:
            methodCode.append(self.__indentStr(f'Args:\n'))

        for i in range(len(params)):
            param: PyutParameter = params[i]
            methodCode.append(self.__indentStr(f'{param.name}:\n', 2))

        # Add others
        returnType: PyutType = pyutMethod.returnType
        if returnType is not None and len(str(returnType)) > 0:
            methodCode.append(self.__indentStr('Returns:\n'))
            methodCode.append(self.__indentStr(f'{returnType}\n', 2))

        return methodCode

    def __addParamToMethodSignature(self, currentCode, paramCode):
        """
        Is smart enough to know if the parameters list is so long it must be indented

        Args:
            currentCode:    Generated 'so far` code
            paramCode:      Generated code for current parameter

        Returns:
            Updated code
        """
        if (len(currentCode) % PyutToPython.MAX_WIDTH) + len(paramCode) > PyutToPython.MAX_WIDTH:  # Width limit
            currentCode += "\n" + self.__indentStr(self.__indentStr(paramCode))
        else:
            currentCode = f'{currentCode}{paramCode}'
        return currentCode

    def __indentStr(self, stringToIndent: str, numTabs: int = 1) -> str:
        """
        Indent one string by one unit

        Args:
            stringToIndent:  string to indent
            numTabs:         number of tabs to insert

        Returns:
            Indented string
        """
        insertedTabs: str = ''
        for x in range(numTabs):
            insertedTabs = f'{insertedTabs}{PyutToPython.SINGLE_TAB}'

        return f'{insertedTabs}{stringToIndent}'

    def __isDunderMethod(self, methodName: str) -> bool:
        """
        Actually, checks to see if a method name already has leading or trailing underscore(s);

        Returns:  `True` if the names starts with underscores else `False`
        """
        isDunder: bool = False
        if methodName.startswith('_') or methodName.startswith('__'):
            isDunder = True
        return isDunder

    def indent(self, listIn: List[str]) -> List[str]:
        """
        Indent every line in listIn by one unit

        Args:
            listIn: Many strings

        Returns:
            A new list with indented strings
        """
        listOut: List[str] = []
        for el in listIn:
            listOut.append(self.__indentStr(str(el)))
        return listOut




from typing import cast
from typing import Dict
from typing import Tuple
from typing import List
from typing import NewType

from logging import Logger
from logging import getLogger

from xml.parsers.expat import ParserCreate
from pyexpat import XMLParserType

from pyutmodel.PyutType import PyutType
from pyutmodel.PyutVisibilityEnum import PyutVisibilityEnum
from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutField import PyutField
from pyutmodel.PyutLinkType import PyutLinkType

from ogl.OglClass import OglClass
from ogl.OglLink import OglLink

from pyutplugins.common.ElementTreeData import ElementTreeData
from pyutplugins.common.LinkMakerMixin import LinkMakerMixin
from pyutplugins.common.PluginTypes import ClassPair
from pyutplugins.common.PluginTypes import ClassTree

from pyutplugins.ExternalTypes import OglClasses
from pyutplugins.ExternalTypes import OglLinks

from pyutplugins.ioplugins.dtd.DTDAttribute import DTDAttribute
from pyutplugins.ioplugins.dtd.DTDElementTypes import DTDElementTypes

DTDElements   = NewType('DTDElements',   Dict[str, Tuple])
DTDAttributes = NewType('DTDAttributes', List[DTDAttribute])


class DTDParser(LinkMakerMixin):

    MODEL_CHILD_ELEMENT_TYPE_INDEX:                int = 0
    MODEL_CHILD_ELEMENT_NAME_INDEX:                int = 2
    MODEL_CHILD_ELEMENT_ADDITIONAL_ELEMENTS_INDEX: int = 3

    MODEL_CHILDREN_INDEX:           int = 3

    def __init__(self):
        """
        Also, I use the pycharm noinspection pragma because I cannot get the correct
        type imported for the parser; I think because the code is 'generated' with
        some kind of C language binder;

        """
        super().__init__()
        self.logger:    Logger                = getLogger(__name__)

        self._elementTypes:   DTDElements    = DTDElements({})
        self._attributes:     DTDAttributes  = DTDAttributes([])
        self._classTree:      ClassTree      = ClassTree({})
        self._links:          OglLinks       = OglLinks([])
        self._oglClasses:     OglClasses     = OglClasses([])

        # noinspection SpellCheckingInspection
        """
        Due to limitations in the Expat library used by pyexpat, the xmlparser instance returned can
        only be used to parse a single XML document.Call ParserCreate for each document to provide unique
        parser instances.
        """
        self._dtdParser: XMLParserType = ParserCreate()

        # noinspection SpellCheckingInspection
        self._dtdParser.StartDoctypeDeclHandler = self.startDocumentTypeHandler
        # noinspection SpellCheckingInspection
        self._dtdParser.ElementDeclHandler      = self.elementHandler
        # noinspection SpellCheckingInspection
        self._dtdParser.AttlistDeclHandler      = self.attributeListHandler
        # noinspection SpellCheckingInspection
        self._dtdParser.EndDoctypeDeclHandler   = self.endDocumentTypeHandler

    def open(self, filename: str) -> bool:
        """

        Args:
            filename:

        Returns:  'True' if opened and parsed correctly else 'False'

        """
        self.logger.info(f'filename: {filename}')

        with open(filename, "r") as dataFile:
            dtdData: str = dataFile.read()
            self._dtdParser.Parse(dtdData)

        return True

    @property
    def classTree(self) -> ClassTree:
        assert len(self._classTree) != 0, 'You should call open first'
        return self._classTree

    @property
    def oglClasses(self) -> OglClasses:
        if len(self._oglClasses) == 0:
            classTree: ClassTree = self.classTree
            for className in classTree.keys():
                elementTreeData: ElementTreeData = classTree[className]
                self._oglClasses.append(elementTreeData.oglClass)

        return self._oglClasses

    @property
    def links(self) -> OglLinks:
        assert len(self._classTree) != 0, 'You should call open first'  # Maybe does not have any links
        return self._links

    def startDocumentTypeHandler(self, docTypeName, sysId, pubId, hasInternalSubset):

        dbgStr: str = f'startDocTypeHandler - {docTypeName=} {sysId=} {pubId=} {hasInternalSubset=}'
        self.logger.info(dbgStr)

    def elementHandler(self, elementName: str, model):
        # noinspection SpellCheckingInspection
        """

        Args:
            elementName:   Element name
            model:  The element content model in (sep,cont,mod) format, where cont is a list of (name,mod) and (sep,cont,mod) tuples.
            ANY content models are represented as None, and EMPTYs as ("",[],"").

            (name , descr , (attribute | attribute-group-ref)*)
        """
        currentLineNumber: int = self._dtdParser.CurrentLineNumber
        self.logger.debug(f'elementHandler - {currentLineNumber:{2}} name: {elementName:{12}} model: {model}')

        self._elementTypes[elementName] = model

    def attributeListHandler(self, eltName, attrName, attrType, attrValue, valType: int):

        dtdAttribute: DTDAttribute = DTDAttribute()

        dtdAttribute.elementName    = eltName
        dtdAttribute.attributeName  = attrName
        dtdAttribute.attributeType  = attrType
        dtdAttribute.attributeValue = attrValue
        dtdAttribute.valueType      = valType

        self.logger.debug(dtdAttribute)

        self._attributes.append(dtdAttribute)

    def endDocumentTypeHandler(self):

        self._classTree = self._createClassTree()
        self.logger.debug(f'elementsTree: {self._classTree}')
        self._addAttributesToClasses()
        self._addLinks()

        self.logger.info(f'attributes: {self._attributes}')

    def _createClassTree(self) -> ClassTree:

        elementsTree: ClassTree = ClassTree({})
        x: int = 50
        y: int = 50

        for eltName in list(self._elementTypes.keys()):

            classPair: ClassPair = self._createClassPair(name=eltName, x=x, y=y)
            pyutClass: PyutClass = classPair.pyutClass
            oglClass:  OglClass  = classPair.oglClass

            elementTreeData: ElementTreeData = ElementTreeData(pyutClass=pyutClass, oglClass=oglClass)

            model = self._elementTypes[eltName]
            # noinspection SpellCheckingInspection
            chillunNames: List[str] = self._getChildElementNames(eltName=eltName, model=model)
            elementTreeData.childElementNames = chillunNames

            elementsTree[eltName] = elementTreeData

            # Carefully, update the graphics layout
            if x < 800:
                x += 80
            else:
                x = 80
                y += 80

        return elementsTree

    def _addLinks(self):

        for className in list(self._classTree.keys()):

            eltTreeData: ElementTreeData = self._classTree[className]

            for associatedClassName in eltTreeData.childElementNames:

                self.logger.info(f'{className} associated with {associatedClassName}')

                parent:      OglClass = eltTreeData.oglClass
                dstTreeData: ElementTreeData = self._classTree[associatedClassName]
                child:       OglClass = dstTreeData.oglClass

                link: OglLink = self.createLink(parent, child, PyutLinkType.AGGREGATION)

                self._links.append(link)

    def _addAttributesToClasses(self):

        for classAttr in self._attributes:
            typedAttr: DTDAttribute   = cast(DTDAttribute, classAttr)
            className: str            = typedAttr.elementName
            treeData: ElementTreeData = self._classTree[className]
            attrName: str             = typedAttr.attributeName
            attrType: str             = typedAttr.attributeType
            attrValue: str            = typedAttr.attributeValue

            pyutField: PyutField = PyutField(name=attrName,
                                             fieldType=PyutType(value=attrType),
                                             defaultValue=attrValue,
                                             visibility=PyutVisibilityEnum.PUBLIC)

            self.logger.info(f'pyutField: {pyutField}')
            pyutClass: PyutClass = treeData.pyutClass
            pyutClass.addField(pyutField)

    def _getChildElementNames(self, eltName, model) -> List[str]:

        self.logger.debug(f'_getChildElementNames - {eltName=}\n {model[0]=}\n {model[1]=}\n {model[2]=}\n {model[3]=}')

        children = model[DTDParser.MODEL_CHILDREN_INDEX]
        self.logger.info(f'children {children}')
        # noinspection SpellCheckingInspection
        chillunNames: List[str] = []
        for child in children:

            self.logger.info(f'eltName: {eltName} - child Length {len(child)}')

            dtdElementType: DTDElementTypes = DTDElementTypes(child[DTDParser.MODEL_CHILD_ELEMENT_TYPE_INDEX])
            childName = child[DTDParser.MODEL_CHILD_ELEMENT_NAME_INDEX]
            self.logger.info(f'eltName: {eltName} child: `{child}` eltType: `{dtdElementType.__repr__()}` childName: `{childName}`')
            additionalElements = child[DTDParser.MODEL_CHILD_ELEMENT_ADDITIONAL_ELEMENTS_INDEX]
            if len(additionalElements) != 0:
                for additionChildren in additionalElements:
                    additionalChildName = additionChildren[DTDParser.MODEL_CHILD_ELEMENT_NAME_INDEX]
                    chillunNames.append(additionalChildName)
            else:
                chillunNames.append(childName)

        self.logger.info(f'Children names: {chillunNames}')
        return chillunNames

    def _createClassPair(self, name: str, x: int, y: int) -> ClassPair:
        """
        Create a pair of classes (pyutClass and oglClass)

        TODO:  This method may belong in a common place for plugin development.

        Args:
            name: Class Name

            x:  x-coordinate on the uml frame  oglClass
            y:  y coordinate on the uml frame  oglClass

        Returns: The class pair
        """
        pyutClass: PyutClass = PyutClass()
        pyutClass.name = name

        oglClass: OglClass = OglClass(pyutClass, 50, 50)
        # To make this code capable of being debugged
        oglClass.SetPosition(x=x, y=y)
        # self.addShape(oglClass, x, y)     # TODO:  adding to UML frame does not belong here

        classPair: ClassPair = ClassPair(pyutClass=pyutClass, oglClass=oglClass)

        return classPair




from typing import cast
from typing import Dict
from typing import List

from logging import Logger
from logging import getLogger

from os import sep as osSep

from pyutmodel.PyutClass import PyutClass

from ogl.OglClass import OglClass

from wx import ICON_ERROR
from wx import OK
from wx import PD_APP_MODAL
from wx import PD_ELAPSED_TIME

from wx import MessageBox
from wx import BeginBusyCursor
from wx import EndBusyCursor
from wx import ProgressDialog

from wx import Yield as wxYield

from pyutplugins.IPluginAdapter import IPluginAdapter
from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface

from pyutplugins.ExternalTypes import OglClasses
from pyutplugins.ExternalTypes import OglLinks
from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import PluginName
from pyutplugins.plugintypes.PluginDataTypes import FormatName
from pyutplugins.plugintypes.PluginDataTypes import PluginDescription

from pyutplugins.plugintypes.ExportDirectoryResponse import ExportDirectoryResponse
from pyutplugins.plugintypes.MultipleFileRequestResponse import MultipleFileRequestResponse
from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat

from pyutplugins.ioplugins.python.PyutToPython import MethodsCodeType
from pyutplugins.ioplugins.python.PyutToPython import PyutToPython
from pyutplugins.ioplugins.python.ReverseEngineerPython2 import ReverseEngineerPython2

FORMAT_NAME:        FormatName        = FormatName("Python File(s)")
PLUGIN_EXTENSION:   PluginExtension   = PluginExtension('py')
PLUGIN_DESCRIPTION: PluginDescription = PluginDescription('Python code generation and reverse engineering')


class IOPython(IOPluginInterface):

    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        # from super class
        self._name    = PluginName('IOPython')
        self._author  = 'Humberto A. Sanchez II'
        self._version = '1.0'
        self._inputFormat  = InputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)
        self._outputFormat = OutputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)

        self._exportDirectoryName: str         = ''
        self._importDirectoryName: str         = ''
        self._filesToImport:       List['str'] = []

        self._readProgressDlg: ProgressDialog = cast(ProgressDialog, None)

    def setImportOptions(self) -> bool:
        """
        We do need to ask for the input file names

        Returns:  'True', we support import
        """
        response: MultipleFileRequestResponse = self.askToImportMultipleFiles(startDirectory=None)
        if response.cancelled is True:
            return False
        else:
            self._importDirectoryName = response.directoryName
            self._filesToImport   = response.fileList

        return True

    def setExportOptions(self) -> bool:
        response: ExportDirectoryResponse = self.askForExportDirectoryName(preferredDefaultPath=None)
        if response.cancelled is True:
            return False
        else:
            self._exportDirectoryName = response.directoryName
            return True

    def read(self) -> bool:
        """

        Returns:
        """
        BeginBusyCursor()
        wxYield()
        status: bool = True
        try:
            reverseEngineer: ReverseEngineerPython2 = ReverseEngineerPython2()

            fileCount: int        = len(self._filesToImport)
            self._readProgressDlg = ProgressDialog('Parsing Files', 'Starting',  parent=None, style=PD_APP_MODAL | PD_ELAPSED_TIME)
            self._readProgressDlg.SetRange(fileCount)

            reverseEngineer.reversePython(directoryName=self._importDirectoryName, files=self._filesToImport, progressCallback=self._readProgressCallback)
            oglClasses: OglClasses = reverseEngineer.oglClasses
            oglLinks:   OglLinks   = reverseEngineer.oglLinks()

            self.logger.warning(f'Classes: {oglClasses}')
            self.logger.warning(f'Links:   {oglLinks}')

            self._layoutUmlClasses(oglClasses=oglClasses)
            self._layoutLinks(oglLinks=oglLinks)
        except (ValueError, Exception) as e:
            self._readProgressDlg.Destroy()
            MessageBox(f'{e}', 'Error', OK | ICON_ERROR)
            status = False
        else:
            self._readProgressDlg.Destroy()
            EndBusyCursor()
            self._pluginAdapter.indicatePluginModifiedProject()
        return status

    def write(self, oglObjects: OglObjects):

        directoryName: str = self._exportDirectoryName

        self.logger.info("IoPython Saving...")
        pyutToPython: PyutToPython = PyutToPython()
        classes:           Dict[str, List[str]] = {}
        generatedClassDoc: List[str]            = pyutToPython.generateTopCode()
        # oglClasses: OglClasses = self._communicator.selectedOglObjects

        for oglClass in oglObjects:
            if isinstance(oglClass, OglClass):
                pyutClass:          PyutClass = oglClass.pyutObject
                generatedStanza:    str       = pyutToPython.generateClassStanza(pyutClass)
                generatedClassCode: List[str] = [generatedStanza]

                clsMethods: MethodsCodeType = pyutToPython.generateMethodsCode(pyutClass)

                # Add __init__ Method
                if PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR in clsMethods:
                    methodCode = clsMethods[PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR]
                    generatedClassCode += methodCode
                    del clsMethods[PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR]

                # Add others methods in order
                for pyutMethod in pyutClass.methods:
                    methodName: str = pyutMethod.name
                    if methodName != PyutToPython.SPECIAL_PYTHON_CONSTRUCTOR:
                        try:
                            otherMethodCode: List[str] = clsMethods[methodName]
                            generatedClassCode += otherMethodCode
                        except (ValueError, Exception, KeyError) as e:
                            self.logger.warning(f'{e}')

                generatedClassCode.append("\n\n")
                # Save into classes dictionary
                classes[pyutClass.name] = generatedClassCode

        # Write class code to a file
        for (className, classCode) in list(classes.items()):
            self._writeClassToFile(classCode, className, directoryName, generatedClassDoc)

        self.logger.info("IoPython done !")

    def _writeClassToFile(self, classCode, className, directory, generatedClassDoc):

        filename: str = f'{directory}{osSep}{str(className)}.py'

        file = open(filename, "w")
        file.writelines(generatedClassDoc)
        file.writelines(classCode)

        file.close()

    def _readProgressCallback(self, currentFileCount: int, msg: str):
        """

        Args:
            currentFileCount:   The current file # we are working pm
            msg:    An updated message
        """

        self._readProgressDlg.Update(currentFileCount, msg)




from typing import cast

from logging import Logger
from logging import getLogger

from time import time

from wx import ICON_ERROR
from wx import OK

from wx import MessageBox

from wx import Yield as wxYield

from miniogl.Shape import Shape

from ogl.OglClass import OglClass
from ogl.OglLink import OglLink
from ogl.OglNote import OglNote

from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.plugininterfaces.ToolPluginInterface import ToolPluginInterface

from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.plugintypes.PluginDataTypes import PluginName

from pyutplugins.toolplugins.orthogonal.DlgLayoutSize import DlgLayoutSize
from pyutplugins.toolplugins.orthogonal.OrthogonalAdapter import LayoutAreaSize
from pyutplugins.toolplugins.orthogonal.OrthogonalAdapter import OglCoordinate
from pyutplugins.toolplugins.orthogonal.OrthogonalAdapter import OglCoordinates
from pyutplugins.toolplugins.orthogonal.OrthogonalAdapter import OrthogonalAdapter
from pyutplugins.toolplugins.orthogonal.OrthogonalAdapterException import OrthogonalAdapterException


class ToolOrthogonalLayoutV2(ToolPluginInterface):

    """
    Version 2 of this plugin.  Does not depend on python-tulip.  Instead, it depends on a homegrown
    version
    """
    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._layoutWidth:  int = 0
        self._layoutHeight: int = 0

        self._name      = PluginName('Orthogonal Layout')
        self._author    = 'Humberto A. Sanchez II'
        self._version   = '2.1'

        self._menuTitle = 'Orthogonal Layout V2'

    def setOptions(self) -> bool:

        with DlgLayoutSize(None) as dlg:
            dlgLayoutSize: DlgLayoutSize = cast(DlgLayoutSize, dlg)
            if dlgLayoutSize.ShowModal() == OK:
                self.logger.warning(f'Retrieved data: layoutWidth: {dlgLayoutSize.layoutWidth} layoutHeight: {dlgLayoutSize.layoutHeight}')
                self._layoutWidth  = dlgLayoutSize.layoutWidth
                self._layoutHeight = dlgLayoutSize.layoutHeight
            else:
                self.logger.warning(f'Cancelled')

        return True

    def doAction(self):
        self._pluginAdapter.getSelectedOglObjects(callback=self._doAction)

    def _doAction(self, selectedObjects: OglObjects):

        try:
            orthogonalAdapter: OrthogonalAdapter = OrthogonalAdapter(umlObjects=selectedObjects)

            layoutAreaSize: LayoutAreaSize = LayoutAreaSize(self._layoutWidth, self._layoutHeight)
            orthogonalAdapter.doLayout(layoutAreaSize)
        except OrthogonalAdapterException as oae:
            MessageBox(f'{oae}', 'Error', OK | ICON_ERROR)
            return

        if orthogonalAdapter is not None:
            self._reLayoutNodes(selectedObjects, orthogonalAdapter.oglCoordinates)
            self._reLayoutLinks(selectedObjects)
            self._pluginAdapter.indicatePluginModifiedProject()

    def _reLayoutNodes(self, umlObjects: OglObjects, oglCoordinates: OglCoordinates):
        """

        Args:
            umlObjects:
        """

        for umlObj in umlObjects:
            if isinstance(umlObj, OglClass) or isinstance(umlObj, OglNote):
                oglName: str = umlObj.pyutObject.name
                oglCoordinate: OglCoordinate = oglCoordinates[oglName]

                self._stepNodes(umlObj, oglCoordinate)
            self._animate()

    def _reLayoutLinks(self, umlObjects: OglObjects):

        for oglObject in umlObjects:
            if isinstance(oglObject, OglLink):
                oglLink: OglLink = cast(OglLink, oglObject)
                oglLink.optimizeLine()
            self._animate()

    def _stepNodes(self, srcShape: Shape, oglCoordinate: OglCoordinate):

        oldX, oldY = srcShape.GetPosition()
        newX: int = oglCoordinate.x
        newY: int = oglCoordinate.y

        self.logger.info(f'{srcShape} - oldX,oldY: ({oldX},{oldY}) newX,newY: ({newX},{newY})')
        #
        srcShape.SetPosition(newX, newY)

    def _animate(self):
        """
        Does an animation simulation
        """
        # umlFrame.Refresh()
        self._pluginAdapter.refreshFrame()
        self.logger.debug(f'Refreshing ...............')
        wxYield()
        t = time()
        while time() < t + 0.05:
            pass




from typing import Dict
from typing import List
from typing import NewType
from typing import Set
from typing import TextIO
from typing import Union
from typing import cast

from logging import Logger
from logging import getLogger

from ogl.OglInterface import OglInterface
from ogl.OglLink import OglLink
from ogl.OglClass import OglClass
from ogl.OglInterface2 import OglInterface2

from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutField import PyutField
from pyutmodel.PyutLinkType import PyutLinkType
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutParameter import PyutParameter
from pyutmodel.PyutType import PyutType
from pyutmodel.PyutVisibilityEnum import PyutVisibilityEnum

from pyutplugins.common.LinkMakerMixin import LinkMakerMixin

Implementors = NewType('Implementors', Set[OglClass])
Extenders    = NewType('Extenders', Set[OglClass])

Links           = Union[OglLink, OglInterface2]
ReversedClasses = NewType('ReversedClasses', Dict[str, OglClass])
ReversedLinks   = NewType('ReversedLinks', List[Links])
SubClassMap     = NewType('SubClassMap', Dict[OglClass, Extenders])
InterfaceMap    = NewType('InterfaceMap', Dict[OglClass, Implementors])

# Constants
CLASS_MODIFIER   = ["public", "protected", "private", "abstract", "final", "static", "strictfp"]
FIELD_MODIFIERS  = ["public", "protected", "private", "static", "final", "transient", "volatile"]
METHOD_MODIFIERS = ["public", "protected", "private", "abstract", "static", "final", "synchronized", "native", "strictfp"]


class JavaReader(LinkMakerMixin):
    """
    TODO:  Figure out how to do associations
    """

    def __init__(self):

        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._classes:       ReversedClasses = ReversedClasses({})
        self._subClassMap:   SubClassMap     = SubClassMap({})
        self._interfaceMap:  InterfaceMap    = InterfaceMap({})
        self._reversedLinks: ReversedLinks   = cast(ReversedLinks, None)

    @property
    def reversedClasses(self) -> ReversedClasses:
        return self._classes

    @property
    def subClassMap(self) -> SubClassMap:
        return self._subClassMap

    @property
    def interfaceMap(self) -> InterfaceMap:
        return self._interfaceMap

    @property
    def reversedLinks(self) -> ReversedLinks:
        if self._reversedLinks is None:
            self._reversedLinks = self._connectLinks()
        return self._reversedLinks

    def parseFile(self, filename: str):
        """
        Creates entries accessible via the reverseClasses property.
        Superclass and subclass relationships are captured in the subClassMap property
        Interface relationships are captured in the interfaceMap property.

        Args:
            filename:  The java file to parse

        """
        # Read the file into a list of strings one for each line
        reader:       TextIO = open(filename, "r")
        temporaryList: List[str] = reader.readlines()
        reader.close()
        #
        # Split file into many words
        wordList: List[str] = []
        for el in temporaryList:
            fraction = el.split()
            wordList += self.__mySplit(fraction)
            wordList.append("\n")

        # parse the file
        currentPos = 0
        try:
            while currentPos < len(wordList):
                # Analyze word
                self.__logMessage(f"\n analyzing word {currentPos}")
                self.__logMessage("***")
                self.__logMessage(f"{wordList}")

                # Read comments
                currentPos = self._readComments(wordList, currentPos)

                # new class ?
                if self._isClassBeginning(wordList, currentPos):
                    currentPos = self._readClass(wordList, currentPos)
                currentPos += 1
        except (ValueError, Exception) as e:
            self.logger.error(f'Error in ReverseJava - {type(e)} {e}')
            raise e

    def _readClass(self, wordList: List[str], currentPos: int):
        """
        Read a class from a list of strings, beginning on a given position.
        This class reads one class from lstFile, from currentLine.

        Args:
            wordList:   list of instructions read from the file to analyze
            currentPos: last read line pointer, OglClass object, class name

        Returns:  Returns the updated current position (read line pointer)
        """
        # Pass modifiers
        while wordList[currentPos] in CLASS_MODIFIER:
            currentPos += 1

        # Read "class"
        currentPos += 1

        # Read class name
        className = wordList[currentPos]
        self.__logMessage(f"Reading className {className}")
        currentPos += 1

        # Create a class object
        self.__addClass(className)

        # Remove end of lines
        currentPos = self.__selectNextAfter(wordList, currentPos, ['\n', ''])

        # Read inheritance and interface parameters
        while wordList[currentPos] in ["extends", "implements"]:
            if wordList[currentPos] == "extends":
                # Next token, points on superClass name
                currentPos += 1

                # Get superclass
                superClass = wordList[currentPos]
                currentPos += 1
                self.__logMessage(f' - superclass={superClass}')

                # Create a class object
                self.__addClass(superClass)
                self.__updateExtendersMap(className, superClass)

            else:  # implements
                shouldExit: bool = False
                self.__logMessage(f"Reading interface... {wordList[currentPos:currentPos + 5]}")
                while not shouldExit:
                    # Next token, points on first interface name
                    currentPos += 1
                    currentPos = self.__selectNextAfter(wordList, currentPos, ['\n', ''])

                    # Get interface name
                    interfaceName = wordList[currentPos]
                    self.__logMessage(f" - interface={interfaceName}")
                    currentPos += 1

                    # Create a class object
                    self.__addClass(interfaceName)
                    self.__updateInterfaceMap(className, interfaceName)

                    # Read comments
                    currentPos = self._readComments(wordList, currentPos)
                    currentPos = self.__selectNextAfter(wordList, currentPos, ['\n', ''])

                    # Exit if no more implementations
                    if wordList[currentPos] != ",":
                        shouldExit = True

            # Remove end of lines
            currentPos = self.__selectNextAfter(wordList, currentPos, ['\n', ''])

        # Read comments
        currentPos = self._readComments(wordList, currentPos)

        # Remove end of lines
        currentPos = self.__selectNextAfter(wordList, currentPos, ['\n', ''])

        # End of class ?
        if wordList[currentPos] == ";":
            return currentPos

        # Remove end of lines
        currentPos = self.__selectNextAfter(wordList, currentPos, ['\n', ''])

        # Read comments
        currentPos = self._readComments(wordList, currentPos)

        # Class beginning ?
        self.__logMessage("lstFile=%s" % wordList[currentPos:currentPos + 5])
        if wordList[currentPos] != "{":
            self.__logMessage(f"DBG class >> {wordList[currentPos]}")
            self.__logMessage(f"Unexpected characters : {wordList[currentPos:currentPos + 5]}")
            self.__logMessage("   exiting class reader !\n")
            return currentPos
        currentPos += 1

        # Read class
        level = 1  # level of indentation
        while level > 0 and currentPos < len(wordList):
            # Read comments
            currentPos = self._readComments(wordList, currentPos)

            # Change level ?
            if wordList[currentPos] == "{":
                level += 1
            elif wordList[currentPos] == "}":
                level -= 1
            elif level == 1:
                (succeeded, aTuple, currentPos) = self._readVariable(wordList, currentPos, True)
                if succeeded:
                    (modifiers, fieldType, names_values) = aTuple
                    self.__addClassFields(className, modifiers, fieldType, names_values)
                else:
                    # Read comments
                    currentPos = self._readComments(wordList, currentPos)

                    # Read method
                    (succeeded, aTuple, currentPos) = self._readMethod(wordList, currentPos, className, True)
                    if succeeded:
                        (modifiers, methodType, name, lstFields) = aTuple
                        self.__addClassMethod(className, modifiers, methodType, name, lstFields)

            # Next token ?
            if level > 0:
                currentPos += 1
        return currentPos

    def _readComments(self, lstFile, currentPos):
        """
        Read all comments

        @param lstFile : list of instructions read from the file to analyze
        @param currentPos : current position in the list
        @author C.Dutoit <dutoitc@hotmail.com>
        @since 1.0
        """
        self.__logMessage("readComments")
        # Set comment terminator
        if lstFile[currentPos][:2] == "//":
            terminator = "\n"
        elif lstFile[currentPos][:2] == "/*":
            terminator = "*/"
        else:  # No comments => return initial position
            return currentPos

        # Pass comments
        currentPos += 1
        while lstFile[currentPos][-len(terminator):] != terminator and currentPos < len(lstFile):
            currentPos += 1
        return currentPos + 1

    def _readMethod(self, lstFile, currentPos, className, returnValues=False):
        """
        Read a method.
        Is considered as function :
        - Something of the following type :
            Method_Declaration ::= Method_Header MethodBody
            Method_Header ::= Method_Modifiers Result_Type Method_Declaration
                              Throws
            ResultType ::= Type or void
            Method_Declaration ::= identifier (Formal_Parameter_List)
            Formal_Parameter_List ::= [Formal_Parameter_List,] Formal_Parameter
            Formal_Parameter ::= final Type Variable_Declaration_Id
            Throws ::= throws Class_Type_List
            Class_Type_List ::= [Class_Type_List,] ClassType
            ...

        @param lstFile : list of instructions read from the file to analyze
        @param currentPos : current position in the list
        @param className : current class name, used to detect constructors
        @param returnValues : bool, indicate if we must return values
                              (plugintypes, ...)
        @return bool : True if a variable start at the given position
                or tuple : (bool, True if succeeded;
                            tuple : (modifiers, type, name, parameters),
                            final position)
                parameters type if list of tuple (type, name, def value)
        @Author C.Dutoit
        @since 1.0
        """
        # Init
        pos = currentPos
        lstModifiers = []

        # Debug info
        self.__logMessage("readMethod, %s..." % (lstFile[currentPos:currentPos + 5]))

        # Pass field modifiers
        while lstFile[pos] in METHOD_MODIFIERS:
            lstModifiers.append(lstFile[pos])
            pos += 1

        # Read type (only for non-constructors)
        if lstFile[pos] != className:
            returnType = lstFile[pos]
            pos += 1
        else:
            returnType = ""

        # Read method name (identifier)
        methodName = lstFile[pos]
        pos += 1

        # valid following ?
        if lstFile[pos] != "(":
            if returnValues:
                return False, None, currentPos
            else:
                return False

        # Read the parameters
        # TODO : int class(int pi=(...)); doesn't work here ! (parenthesis)
        parameters = []
        pos += 1
        self.__logMessage("*******************************")
        self.__logMessage("Reading parameters %s" % lstFile[pos:pos + 10])
        while lstFile[pos] != ")":
            # Get parameter
            paramType = lstFile[pos]
            pos += 1
            name = lstFile[pos]
            pos += 1
            self.__logMessage("type=%s, name=%s" % (type, name))

            # Read default value
            defaultValue = ""
            if lstFile[pos] == "=":
                pos += 1
                while lstFile[pos] not in [")", ","]:
                    defaultValue += lstFile[pos] + " "
                    pos += 1
                if lstFile[pos] == ",":
                    pos += 1
                self.__logMessage(f"={defaultValue}")
            if defaultValue == "":
                defaultValue = ''

            # Pass \n ?
            pos = self.__selectNextAfter(lstFile, pos, ["\n"])

            # ',' => new parameter => next position
            if lstFile[pos] == ",":
                pos += 1

            # Pass \n ?
            pos = self.__selectNextAfter(lstFile, pos, ["\n"])

            # Append to parameters list
            parameters.append((paramType, name, defaultValue))
        # pos = self._readParagraph(lstFile, pos, "(", ")")
        self.__logMessage("*******************************")
        pos += 1

        # Pass \n ?
        pos = self.__selectNextAfter(lstFile, pos, ["\n"])

        # pass throws ?
        if lstFile[pos] == "throws":
            pos += 1
            while lstFile[pos][-1] not in [";", "{"]:
                pos += 1

        # End of function ?
        if lstFile[pos][-1] == ";":
            if returnValues:
                return True, (lstModifiers, returnType, methodName, parameters), pos
            else:
                return True

        # The following is not a function corps ? => return
        if lstFile[pos] != "{":
            if returnValues:
                return False, None, currentPos

        # Read function corps
        pos = self.__readParagraph(lstFile, pos, "{", "}")
        if returnValues:
            return True, (lstModifiers, returnType, methodName, parameters), pos
        else:
            return True

    def _readVariable(self, lstFile, currentPos, returnValues=False):
        """
        Test if the current position starts a variable.
        Is considered as variable :
        - Something of the following type :
           Field_Declaration    ::= Field_Modifiers Type Variable_Declaration;
           Variable_Declaration ::= [Variable_Declaration,] Variable_Declaration
           Variable_Declaration  ::= Variable_Declaration_ID
                                    [ = Variable_Initializer ]
           Variable_Initializer ::= ... etc

        @param lstFile : list of instructions read from the file to analyze
        @param currentPos : current position in the list
        @param returnValues : bool, indicate if we must return values
                              (plugintypes, ...)
        @return bool : True if a variable start at the given position
                or tuple : (bool, True if succeeded;
                            tuple : (modifiers, type, tuple: (name, values));
                            final position)
        @Author C.Dutoit
        @since 1.0
        """
        # Init
        pos = currentPos
        lstModifiers = []
        lstNames_Values = []

        # Pass field modifiers
        while lstFile[pos] in FIELD_MODIFIERS:
            lstModifiers.append(lstFile[pos])
            pos += 1

        # Read type
        theType = lstFile[pos]
        pos += 1

        # Read first variable declaration
        lstNames_Values.append((lstFile[pos], None))
        self.__logMessage(f"Adding name {lstFile[pos]}, following {lstFile[pos:pos + 2]}")
        pos += 1

        # valid following ?
        if lstFile[pos] not in [",", "=", ";"]:
            if returnValues:
                return False, None, currentPos
            else:
                return False

        # Continue reading...
        while pos < len(lstFile):
            if lstFile[pos] == ",":
                pos += 1
                # lstModifiers.append(lstFile[pos])
                lstNames_Values.append((lstFile[pos], None))
                pos += 1
            elif lstFile[pos] == ";":
                if returnValues:
                    return True, (lstModifiers, theType, lstNames_Values), pos
                else:
                    return True
            elif lstFile[pos] == "=":
                # TODO : change this; pass values
                pos += 1
                while lstFile[pos] != ";":
                    pos += 1
            else:
                if returnValues:
                    return False, None, currentPos
                else:
                    return False

            # TODO : this does not support int a=8, b=4; ('=')

        if returnValues:
            return False, None, currentPos
        else:
            return False

    def _isClassBeginning(self, lstFile, currentPos):
        """
        Return True if the specified line is a class beginning

        @param lstFile : list of instructions read from the file to analyze
        @param currentPos : current position in the list
        @return bool : True if the specified line do begin a class
        @author C.Dutoit
        @since 1.0
        """
        lstUsedCM = []  # List of used class modifiers
        # Evaluate each argument
        pos = currentPos
        while pos < len(lstFile):
            el = lstFile[pos]
            # Is current argument a class modifier ?
            if el in CLASS_MODIFIER:
                # if not modified, add it as modifiers list
                if not (el in lstUsedCM):
                    lstUsedCM.append(el)
                else:  # Already used => not a valid class beginning
                    # TODO : print warning ?
                    return False
            elif el == "class":  # class token => this is a class
                return True
            elif el == "interface":  # interface token => take it as a class
                return True
            else:  # unacceptable token => not a class beginning
                return False
            pos += 1
        return False

    def __addClass(self, className: str) -> OglClass:
        """
        Add a class to the dictionary of classes

        Args:
            className: Name of the class to be added

        Returns: OglClass instance for the class

        """
        # If the class name exists already, return the instance
        if className in self._classes:
            return self._classes[className]

        # Create the class
        pc: PyutClass = PyutClass(className)  # A new PyutClass
        po: OglClass = OglClass(pc)  # A new OglClass

        self._classes[className] = po

        return po

    def _connectLinks(self) -> ReversedLinks:

        reversedLinks: ReversedLinks = ReversedLinks([])

        reversedLinks = self._connectSubClasses(reversedLinks)
        reversedLinks = self._connectInterfaces(reversedLinks)

        return reversedLinks

    def _connectSubClasses(self, reversedLinks: ReversedLinks) -> ReversedLinks:

        for parent in self._subClassMap.keys():
            self.logger.info(f'BaseClass: {parent}')
            extenders: Extenders = self._subClassMap[parent]
            for child in extenders:
                self.logger.info(f'{parent} <--- {child}')

                oglLink: OglLink = self.createLink(src=child, dst=parent, linkType=PyutLinkType.INHERITANCE)
                reversedLinks.append(oglLink)

        return reversedLinks

    def _connectInterfaces(self, reversedLinks: ReversedLinks) -> ReversedLinks:

        self.logger.info('Connect interfaces')
        for interface in self._interfaceMap.keys():
            implementors: Implementors = self._interfaceMap[interface]
            for implementor in implementors:
                oglInterface: OglInterface = self.createInterfaceLink(src=implementor, dst=interface)
                reversedLinks.append(oglInterface)

        return reversedLinks

    def __addClassFields(self, className, modifiers, fieldType, names_values):
        """
        Add fields to a class
        Note : Default value can be None. (see names_values)

        @param String className : Name of the class to be added
        @param String modifiers : Fields modifiers
        @param String fieldType      : Fields type
        @param tuple names_values : tuple of (name, default values)

        """
        # Get class fields
        po: OglClass = self._classes[className]
        pc: PyutClass = cast(PyutClass, po.pyutObject)
        classFields = pc.fields

        # TODO fix this crazy code to use constructor and catch exception on bad input
        # Get visibility
        visibility: PyutVisibilityEnum = PyutVisibilityEnum.PUBLIC
        if "private" in modifiers:
            visibility = PyutVisibilityEnum.PRIVATE
        elif "protected" in modifiers:
            visibility = PyutVisibilityEnum.PROTECTED
        elif "public" in modifiers:
            visibility = PyutVisibilityEnum.PUBLIC

        # Add all
        for (name, value) in names_values:
            pyutType: PyutType = PyutType(fieldType)
            classFields.append(PyutField(name, pyutType, value, visibility))

    def __addClassMethod(self, className, modifiers, returnType, name, lstFields):
        """
        Add a method to a class
        Parameters example:
        ```java
            static private int getName(int a, long b)
            - className = name of the class which own this method
            - static private => modifiers
            - int            => return type
            - getName        => name
            - int a, long b  => lstFields => ["int a", "long b"]
        ```
        Note : Default value can be None. (see names_values)

        Args:
            className:  Name of the class to add
            modifiers:  Method modifiers
            returnType: Method returnType
            name:       Method name
            lstFields:  List of string lstFields : Method fields
        """

        self.__logMessage("Adding method %s for class %s" % (name, className))
        self.__logMessage("(modifiers=%s; returnType=%s)" % (modifiers, returnType))
        # Get class fields
        po: OglClass = self._classes[className]
        pc: PyutClass = cast(PyutClass, po.pyutObject)

        # TODO fix this crazy code to use constructor and catch exception on bad input
        # Get visibility
        visibility: PyutVisibilityEnum = PyutVisibilityEnum.PUBLIC
        if "private" in modifiers:
            visibility = PyutVisibilityEnum.PRIVATE
        elif "protected" in modifiers:
            visibility = PyutVisibilityEnum.PROTECTED
        elif "public" in modifiers:
            visibility = PyutVisibilityEnum.PUBLIC

        # Add method
        methods = pc.methods

        if returnType == '\n' or returnType == '' or returnType == 'void' or returnType is None:
            pm = PyutMethod(name, visibility)
        else:
            retType: PyutType = PyutType(returnType)
            pm = PyutMethod(name, visibility, retType)

        for (paramType, name, defaultValue) in lstFields:
            param = PyutParameter(name, paramType, defaultValue)
            pm.addParameter(param)
        methods.append(pm)

    def __readParagraph(self, lstFile, currentPos, paragraphStart, paragraphStop):
        """
        Read a paragraph; Handle the sublevel;
        Read a paragraph. A paragraph is limited by two specific tokens.
        As a paragraph can include itself a paragraph, this function
        does read sub-paragraphs.

        @param lstFile : list of instructions read from the file to analyze
        @param currentPos : current position in the list
        @param paragraphStart : token which identify the beginning of the
                                paragraph
        @param paragraphStop  : token which identify the end of the paragraph
        @return end of paragraph position
        @Author C.Dutoit
        @since 1.1.2.5
        """
        # We must begin on the paragraph start token
        if lstFile[currentPos] != paragraphStart:
            return currentPos

        # Init
        level = 1

        # Read paragraph
        while level > 0:
            currentPos += 1
            if lstFile[currentPos] == paragraphStart:
                level += 1
            elif lstFile[currentPos] == paragraphStop:
                level -= 1
        return currentPos

    def __selectNextAfter(self, lstFile, currentPos, lstElements):
        """
        Select next position after an End-Of-Line

        @param lstFile : list of instructions read from the file to analyze
        @param currentPos : current position in the list
        @param lstElements : Elements to pass
        @return Next position after EOL
        @Author C.Dutoit
        @since 1.0
        """
        while (currentPos < len(lstFile)) and (lstFile[currentPos] in lstElements):
            currentPos += 1
        return currentPos

    def __mySplit(self, lstIn):
        """
        Do more splitting on a list of String.
        Split elements like [")};"]     into    ¨[")", "}", ";"]

        or

        ["Humberto;"]   into   ["Humberto", ";"]


        Args:
            lstIn: list of String to be split

        Returns: the split list
        """
        TO_BE_SPLIT = ['{', '}', ';', ')', '(', ',']
        lstOut = []
        # Split each element
        for el in lstIn:
            pos = 0
            while len(el) > 0 and pos < len(el):
                if el[pos] in TO_BE_SPLIT:
                    if pos > 0:
                        lstOut.append(el[:pos])
                    lstOut.append(el[pos])
                    el = el[pos + 1:]
                    pos = 0
                else:
                    pos += 1

            # Add el
            if len(el) > 0:
                lstOut.append(el)

        # Return split list
        return lstOut

    def __updateExtendersMap(self, className: str, superClass: str):
        """
        Update the map of superclasses (base classes) and the classes that extend
        it (subclasses)
        Args:
            className:  The original class
            superClass: The name of the super class
        """
        oglSuperClass: OglClass = self._classes[superClass]
        oglClass:      OglClass = self._classes[className]

        if oglSuperClass in self._subClassMap:
            extenders: Extenders = self._subClassMap[oglSuperClass]
            extenders.add(oglClass)
        else:
            extenders = Extenders(set())
            extenders.add(oglClass)

        self.logger.debug(f'Make extender entry for {superClass} - subclass: {className}')
        self._subClassMap[oglSuperClass] = extenders

    def __updateInterfaceMap(self, className: str, interfaceName: str):

        interfaceClass: OglClass = self._classes[interfaceName]
        implementor:    OglClass = self._classes[className]

        if interfaceClass in self._interfaceMap:
            implementors: Implementors = self._interfaceMap[interfaceClass]
            implementors.add(implementor)
        else:
            implementors = Implementors(set())
            implementors.add(implementor)

        self.logger.debug(f'Class {className} implements {interfaceName}')
        self._interfaceMap[interfaceClass] = implementors

    def __logMessage(self, theMessage: str):
        """
        Probably not a correct implementation but that is what I got !!
        Args:
            theMessage:

        """
        self.logger.debug(f"{theMessage}")




from typing import cast
from typing import List

from logging import Logger
from logging import getLogger

from wx import EVT_BUTTON
from wx import EVT_CHOICE
from wx import EVT_CLOSE
from wx import EVT_MOTION
from wx import FD_CHANGE_DIR
from wx import FD_OVERWRITE_PROMPT
from wx import FD_SAVE
from wx import ID_CANCEL
from wx import ID_OK
from wx import TE_READONLY

from wx import Button
from wx import Choice
from wx import CommandEvent
from wx import FileDialog
from wx import TextCtrl
from wx import MouseEvent

from wx import Yield as wxYield

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

from pyutplugins.common.ui.BaseEditDialog import BaseEditDialog

from pyutplugins.ioplugins.wximage.WxImageFormat import WxImageFormat

from pyutplugins.preferences.PluginPreferences import PluginPreferences


class DlgWxImageOptions(BaseEditDialog):

    def __init__(self, parent):

        super().__init__(parent, title='Native Image Generation Options')

        self.logger: Logger = getLogger(__name__)

        self._fileSelectBtn:     Button   = cast(Button, None)
        self._selectedFile:      TextCtrl = cast(TextCtrl, None)
        self._imageFormatChoice: Choice   = cast(Choice, None)

        self._outputFileName: str           = PluginPreferences().wxImageFileName
        self._imageFormat:    WxImageFormat = WxImageFormat.PNG

        self._layoutFileSelection(parent=self.GetContentsPane())
        self._layoutImageFormatChoice(parent=self.GetContentsPane())

        self._layoutStandardOkCancelButtonSizer()

        self._bindEventHandlers()

        self.Bind(EVT_BUTTON, self._onOk, id=ID_OK)
        self.Bind(EVT_CLOSE, self._onClose, id=ID_CANCEL)

    @property
    def imageFormat(self) -> WxImageFormat:
        return self._imageFormat

    @imageFormat.setter
    def imageFormat(self, newFormat: WxImageFormat):
        self._imageFormat = newFormat

    @property
    def outputFileName(self) -> str:
        return self._outputFileName

    @outputFileName.setter
    def outputFileName(self, newName: str):
        self._outputFileName = newName

    def _bindEventHandlers(self):

        self.Bind(EVT_BUTTON, self._onFileSelectClick,   self._fileSelectBtn)
        self.Bind(EVT_CHOICE, self._onImageFormatChoice, self._imageFormatChoice)
        #
        self._selectedFile.Bind(EVT_MOTION, self._fileSelectionMotion, self._selectedFile)

    def _fileSelectionMotion(self, event: MouseEvent):

        ctrl: TextCtrl = event.GetEventObject()

        tip = ctrl.GetToolTip()
        tip.SetTip(self._outputFileName)

    # noinspection PyUnusedLocal
    def _onFileSelectClick(self, event: CommandEvent):

        self.logger.warning(f'File Select Click')
        wxYield()

        fmtSelIdx:    int = self._imageFormatChoice.GetCurrentSelection()
        outputFormat: str = self._imageFormatChoice.GetString(fmtSelIdx)

        dlg: FileDialog = FileDialog(self,
                                     message='Choose the export file name',
                                     defaultFile=self._outputFileName,
                                     style=FD_SAVE | FD_OVERWRITE_PROMPT | FD_CHANGE_DIR
                                     )
        if dlg.ShowModal() == ID_OK:
            wxYield()
            path:     str = dlg.GetPath()
            fileName: str = dlg.GetFilename()

            self._selectedFile.SetValue(fileName)       # for simple viewing
            self._selectedFile.SetModified(True)
            self._outputFileName = path                 # for actual us

        dlg.Destroy()

    def _onImageFormatChoice(self, event: CommandEvent):

        ctrl:      Choice = event.GetEventObject()
        idx:       int    = ctrl.GetCurrentSelection()
        newValue:  str    = ctrl.GetString(idx)

        newFormat: WxImageFormat = WxImageFormat(newValue)

        self._imageFormat = newFormat

    def _layoutFileSelection(self, parent: SizedPanel):

        box: SizedStaticBox = SizedStaticBox(parent, -1, "Output Filename")
        box.SetSizerProps(expand=True, proportion=1)

        currentFile: str = self._outputFileName

        sizedPanel: SizedPanel = SizedPanel(box)
        sizedPanel.SetSizerType('horizontal')

        self._fileSelectBtn = Button(sizedPanel, label="&Select",)
        self._selectedFile  = TextCtrl(sizedPanel, value=currentFile, style=TE_READONLY)

        self._selectedFile.SetToolTip(currentFile)

    def _layoutImageFormatChoice(self, parent: SizedPanel):

        imageChoices: List[str] = [WxImageFormat.PNG.value, WxImageFormat.JPG.value, WxImageFormat.BMP.value, WxImageFormat.TIFF.value]

        box: SizedStaticBox = SizedStaticBox(parent, -1, "Image Format")
        box.SetSizerProps(expand=True, proportion=1)

        self._imageFormatChoice = Choice(box, choices=imageChoices)




from typing import cast
from typing import List

from logging import Logger
from logging import getLogger

from time import localtime
from time import strftime

from pyumldiagrams.BaseDiagram import BaseDiagram
from pyumldiagrams.Definitions import ClassDefinition
from pyumldiagrams.Definitions import ClassDefinitions
from pyumldiagrams.Definitions import DefinitionType
from pyumldiagrams.Definitions import DisplayMethodParameters
from pyumldiagrams.Definitions import LinePositions
from pyumldiagrams.Definitions import MethodDefinition
from pyumldiagrams.Definitions import Methods
from pyumldiagrams.Definitions import ParameterDefinition
from pyumldiagrams.Definitions import Parameters
from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import Size
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import UmlLineDefinitions
from pyumldiagrams.Definitions import LineType

from pyumldiagrams.image.ImageDiagram import ImageDiagram

from pyumldiagrams.pdf.PdfDiagram import PdfDiagram

from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutDisplayParameters import PyutDisplayParameters
from pyutmodel.PyutLink import PyutLink
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutVisibilityEnum import PyutVisibilityEnum
from pyutmodel.PyutLinkType import PyutLinkType

from miniogl.AnchorPoint import AnchorPoint
from miniogl.ControlPoint import ControlPoint

from ogl.OglClass import OglClass
from ogl.OglLink import OglLink

from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.ioplugins.pdf.ImageFormat import ImageFormat
from pyutplugins.ioplugins.pdf.ImageOptions import ImageOptions


class OglToPyUmlDefinition:

    INHERITANCE_DESTINATION_POSITION_NUDGE_FACTOR: int = 1

    def __init__(self, imageOptions: ImageOptions, dpi: int = 0, pyutVersion: str = '', pluginVersion: str = ''):
        """

        Args:
            imageOptions: Lots of information on how to draw the diagram

            dpi:  Dots per inch;  Only used in PDF generation;  Image generation is in pixels

            pyutVersion:  Information for header

            pluginVersion:  Information for header
        """

        self.logger:              Logger             = getLogger(__name__)
        self._classDefinitions:   ClassDefinitions   = []
        self._umlLineDefinitions: UmlLineDefinitions = []

        today: str = strftime("%d %b %Y %H:%M:%S", localtime())
        headerText: str = f'Pyut Version {pyutVersion} Plugin Version {pluginVersion} - {today}'

        fqFileName:  str         = imageOptions.outputFileName
        imageFormat: ImageFormat = imageOptions.imageFormat
        #
        # TODO use plugin preferences
        #
        # imageLibShowParameters: DisplayMethodParameters = self.__toImageLibraryEnum(self._prefs.showParameters)
        imageLibShowParameters: DisplayMethodParameters = self.__toImageLibraryEnum(True)
        if imageFormat == ImageFormat.PDF:
            self._diagram: BaseDiagram = PdfDiagram(fileName=fqFileName, dpi=dpi, headerText=headerText, docDisplayMethodParameters=imageLibShowParameters)
        else:
            self._diagram = ImageDiagram(fileName=fqFileName, headerText=headerText)   # TODO use image size from new method signature)

    def toClassDefinitions(self, oglObjects: OglObjects):

        classDefinitions: ClassDefinitions = []

        for oglObject in oglObjects:

            umlObject: OglClass  = cast(OglClass, oglObject)
            if not isinstance(umlObject, OglClass):
                continue

            pyutClass: PyutClass = cast(PyutClass, umlObject.pyutObject)

            x, y = umlObject.GetPosition()
            w, h = umlObject.GetSize()
            position: Position = Position(x=x, y=y)
            size:     Size     = Size(width=int(w), height=int(h))

            classDefinition: ClassDefinition = ClassDefinition(name=pyutClass.name, position=position, size=size)

            if pyutClass.displayParameters is PyutDisplayParameters.DISPLAY:
                classDefinition.displayMethodParameters = DisplayMethodParameters.DISPLAY
            else:
                classDefinition.displayMethodParameters = DisplayMethodParameters.DO_NOT_DISPLAY

            classDefinition = self.__addClassDiagramDisplayPreferences(pyutClass=pyutClass, classDefinition=classDefinition)

            self._addMethods(classDefinition=classDefinition, pyutClass=pyutClass)
            self._diagram.drawClass(classDefinition=classDefinition)
            classDefinitions.append(classDefinition)

        self._classDefinitions = classDefinitions

    def layoutLines(self, oglObjects: OglObjects):

        umlLineDefinitions: UmlLineDefinitions = []

        for umlObject in oglObjects:

            if not isinstance(umlObject, OglLink):
                continue
            oglLink: OglLink = cast(OglLink, umlObject)

            pyutLink:    PyutLink = oglLink.pyutObject
            umlLinkType: PyutLinkType = pyutLink.linkType
            lineType:    LineType = self._toPyUmlLineType(umlLinkType)

            linePositions: LinePositions  = self._toPyUmlPositions(oglLink, umlLinkType)
            self.logger.debug(f'{lineType=} {linePositions=}')

            line:    UmlLineDefinition = UmlLineDefinition(lineType=lineType, linePositions=linePositions)

            self._diagram.drawUmlLine(lineDefinition=line)
            umlLineDefinitions.append(line)

        self._umlLineDefinitions = umlLineDefinitions

    def write(self):
        self._diagram.write()

    def _toPyUmlLineType(self, umlLinkType) -> LineType:

        if umlLinkType == PyutLinkType.INHERITANCE:
            lineType: LineType = LineType.Inheritance
        elif umlLinkType == PyutLinkType.COMPOSITION:
            lineType = LineType.Composition
        elif umlLinkType == PyutLinkType.AGGREGATION:
            lineType = LineType.Aggregation
        else:
            lineType = LineType.Association   # This won't happen yet

        return lineType

    def _toPyUmlPositions(self, oglLink, pyutLinkType: PyutLinkType) -> LinePositions:

        if pyutLinkType == PyutLinkType.INHERITANCE:
            srcAnchor:  AnchorPoint = oglLink.sourceAnchor
            destAnchor: AnchorPoint = oglLink.destinationAnchor
        else:
            srcAnchor  = oglLink.destinationAnchor
            destAnchor = oglLink.sourceAnchor

        srcX,  srcY  = srcAnchor.GetPosition()
        destX, destY = destAnchor.GetPosition()

        sourcePosition:      Position = Position(x=srcX, y=srcY)
        destinationPosition: Position = Position(x=destX, y=destY)

        bends: List[ControlPoint] = oglLink.GetControlPoints()
        if bends is None or len(bends) == 0:
            linePositions: LinePositions = [sourcePosition, destinationPosition]
        else:
            linePositions = [sourcePosition]
            for cp in bends:
                bend: ControlPoint = cast(ControlPoint, cp)
                self.logger.debug(f'{bend:}')

                bendX, bendY = bend.GetPosition()
                bendPosition: Position = Position(x=bendX, y=bendY)
                linePositions.append(bendPosition)

            linePositions.append(destinationPosition)

        return linePositions

    def _addMethods(self, classDefinition: ClassDefinition, pyutClass: PyutClass) -> ClassDefinition:

        methods: Methods = []
        for method in pyutClass.methods:

            pyutMethod: PyutMethod = cast(PyutMethod, method)

            methodDef: MethodDefinition = MethodDefinition(name=pyutMethod.name)

            methodDef.visibility = self.__toDefinitionType(pyutMethod.visibility)
            methodDef.returnType = pyutMethod.returnType.value

            self.__addParameters(methodDefinition=methodDef, pyutMethod=pyutMethod)
            methods.append(methodDef)

        classDefinition.methods = methods
        return classDefinition

    def __addParameters(self, methodDefinition: MethodDefinition, pyutMethod: PyutMethod) -> MethodDefinition:

        parameters: Parameters = []
        for parameter in pyutMethod.parameters:

            paramDef: ParameterDefinition = ParameterDefinition(name=parameter.name)
            paramDef.parameterType = parameter.type.value
            paramDef.defaultValue  = parameter.defaultValue

            parameters.append(paramDef)

        methodDefinition.parameters = parameters
        self.logger.debug(f'{methodDefinition.name=}  {parameters=}')
        return methodDefinition

    def __toDefinitionType(self, visibility: PyutVisibilityEnum) -> DefinitionType:

        if visibility == PyutVisibilityEnum.PUBLIC:
            return DefinitionType.Public
        elif visibility == PyutVisibilityEnum.PRIVATE:
            return DefinitionType.Private
        elif visibility == PyutVisibilityEnum.PROTECTED:
            return DefinitionType.Protected
        else:
            assert False, 'Unknown UML Visibility type.  Probably developer error'

    def __addClassDiagramDisplayPreferences(self, pyutClass: PyutClass, classDefinition: ClassDefinition) -> ClassDefinition:

        classDefinition.displayMethods    = pyutClass.showMethods
        classDefinition.displayFields     = pyutClass.showFields

        return classDefinition

    def __toImageLibraryEnum(self, showParameters: bool) -> DisplayMethodParameters:

        if showParameters is True:
            return DisplayMethodParameters.DISPLAY
        else:
            return DisplayMethodParameters.DO_NOT_DISPLAY




from typing import cast

from logging import Logger
from logging import getLogger

from wx import CANCEL
from wx import CENTRE
from wx import ICON_QUESTION
from wx import YES
from wx import YES_NO

from wx import MessageBox

from wx import Yield as wxYield

from oglio.Reader import Reader
from oglio.Writer import Writer
from oglio.Types import OglProject
from oglio.Types import OglDocument
from oglio.Types import OglDocuments
from oglio.Types import OglDocumentTitle

from pyutplugins.IPluginAdapter import IPluginAdapter
from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface

from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat
from pyutplugins.plugintypes.PluginDataTypes import FormatName
from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import PluginName
from pyutplugins.plugintypes.SingleFileRequestResponse import SingleFileRequestResponse

from pyutplugins.ExternalTypes import OglClasses
from pyutplugins.ExternalTypes import OglLinks
from pyutplugins.ExternalTypes import OglSDInstances
from pyutplugins.ExternalTypes import OglSDMessages
from pyutplugins.ExternalTypes import OglTexts
from pyutplugins.ExternalTypes import OglActors
from pyutplugins.ExternalTypes import OglNotes
from pyutplugins.ExternalTypes import OglObjects
from pyutplugins.ExternalTypes import OglUseCases
from pyutplugins.ExternalTypes import PluginDocument
from pyutplugins.ExternalTypes import PluginDocumentType
from pyutplugins.ExternalTypes import PluginDocumentTitle
from pyutplugins.ExternalTypes import PluginProject

FORMAT_NAME:        FormatName        = FormatName("XML")
PLUGIN_EXTENSION:   PluginExtension   = PluginExtension('xml')
PLUGIN_DESCRIPTION: PluginDescription = PluginDescription('Pyut XML File')


class IOXml(IOPluginInterface):

    def __init__(self, pluginAdapter: IPluginAdapter):

        self.logger: Logger = getLogger(__name__)

        super().__init__(pluginAdapter=pluginAdapter)

        # from super class
        self._name    = PluginName('IOXml')
        # noinspection SpellCheckingInspection
        self._author  = "Humberto A. Sanchez II"
        self._version = '2.0'
        self._inputFormat  = InputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)
        self._outputFormat = OutputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)

        self._prettyPrint:  bool = True
        self._fileToExport: str  = ''
        self._fileToImport: str  = ''

        self._requireActiveFrame = False
        self._requireSelection   = False

    def setImportOptions(self) -> bool:
        """
        We do need to ask for the input file names

        Returns:  'True', we support import
        """
        response: SingleFileRequestResponse = self.askForFileToImport(startDirectory=None)
        if response.cancelled is True:
            return False
        else:
            self._fileToImport   = response.fileName

        return True

    def setExportOptions(self) -> bool:
        """
        Prepare the export.

        Returns:
        """
        prettPrintAnswer: int = MessageBox("Do you want pretty xml ?", "Export option", style=YES_NO | CANCEL | CENTRE | ICON_QUESTION)
        if prettPrintAnswer is YES:
            self._prettyPrint = True
            response: SingleFileRequestResponse = self.askForFileToExport()
            if response.cancelled is True:
                exportAnswer: bool = False
            else:
                self._fileToExport = response.fileName
                exportAnswer = True
        else:
            exportAnswer = False

        return exportAnswer

    def read(self) -> bool:
        reader: Reader = Reader()

        oglProject: OglProject = reader.readXmlFile(fqFileName=self._fileToImport)

        pluginProject: PluginProject = PluginProject()

        pluginProject.fileName    = self._fileToImport
        pluginProject.projectName = PluginProject.toProjectName(fqFilename=self._fileToImport)
        pluginProject.version     = oglProject.version
        pluginProject.codePath    = oglProject.codePath

        oglDocuments: OglDocuments = oglProject.oglDocuments
        for oglDocument in oglDocuments.values():
            pluginDocument: PluginDocument = PluginDocument()
            pluginDocument.documentType    = PluginDocumentType.toEnum(oglDocument.documentType)
            pluginDocument.documentTitle   = PluginDocumentTitle(oglDocument.documentTitle)
            pluginDocument.scrollPositionX = oglDocument.scrollPositionX
            pluginDocument.scrollPositionY = oglDocument.scrollPositionY
            pluginDocument.pixelsPerUnitX  = oglDocument.pixelsPerUnitX
            pluginDocument.pixelsPerUnitY  = oglDocument.pixelsPerUnitY
            pluginDocument.oglClasses      = cast(OglClasses,     oglDocument.oglClasses)
            pluginDocument.oglLinks        = cast(OglLinks,       oglDocument.oglLinks)
            pluginDocument.oglNotes        = cast(OglNotes,       oglDocument.oglNotes)
            pluginDocument.oglTexts        = cast(OglTexts,       oglDocument.oglTexts)
            pluginDocument.oglActors       = cast(OglActors,      oglDocument.oglActors)
            pluginDocument.oglUseCases     = cast(OglUseCases,    oglDocument.oglUseCases)
            pluginDocument.oglSDInstances  = cast(OglSDInstances, oglDocument.oglSDInstances)
            pluginDocument.oglSDMessages   = cast(OglSDMessages,  oglDocument.oglSDMessages)

            pluginProject.pluginDocuments[pluginDocument.documentTitle] = pluginDocument

        self._pluginAdapter.loadProject(pluginProject=pluginProject)
        wxYield()
        self._pluginAdapter.indicatePluginModifiedProject()
        return True

    def write(self, oglObjects: OglObjects):
        """
        Ignoring the input objects since they are from a single frame
        Args:
            oglObjects:
        """

        self._pluginAdapter.requestCurrentProject(callback=self._currentProjectCallback)

    def _currentProjectCallback(self, pluginProject: PluginProject):

        import oglio
        assert isinstance(pluginProject, PluginProject)

        oglProject: OglProject = OglProject()

        oglProject.version  = pluginProject.version
        oglProject.codePath = pluginProject.codePath
        oglProject.fileName = pluginProject.fileName

        for document in pluginProject.pluginDocuments.values():
            pluginDocument: PluginDocument = cast(PluginDocument, document)
            oglDocument:    OglDocument    = OglDocument()

            oglDocument.scrollPositionX = pluginDocument.scrollPositionX
            oglDocument.scrollPositionY = pluginDocument.scrollPositionY
            oglDocument.pixelsPerUnitX  = pluginDocument.pixelsPerUnitX
            oglDocument.pixelsPerUnitY  = pluginDocument.pixelsPerUnitY
            oglDocument.documentTitle   = OglDocumentTitle(pluginDocument.documentTitle)
            oglDocument.documentType    = pluginDocument.documentType.value
            oglDocument.oglClasses      = cast(oglio.Types.OglClasses, pluginDocument.oglClasses)
            oglDocument.oglLinks        = cast(oglio.Types.OglLinks, pluginDocument.oglLinks)
            oglDocument.oglNotes        = cast(oglio.Types.OglNotes, pluginDocument.oglNotes)
            oglDocument.oglTexts        = cast(oglio.Types.OglTexts, pluginDocument.oglTexts)
            oglDocument.oglActors       = cast(oglio.Types.OglActors, pluginDocument.oglActors)
            oglDocument.oglUseCases     = cast(oglio.Types.OglUseCases, pluginDocument.oglUseCases)
            oglDocument.oglSDMessages   = cast(oglio.Types.OglSDMessages, pluginDocument.oglSDMessages)
            oglDocument.oglSDInstances  = cast(oglio.Types.OglSDInstances, pluginDocument.oglSDInstances)

            self.logger.debug(f'{oglDocument=}')

            oglProject.oglDocuments[oglDocument.documentTitle] = oglDocument

        writer:     Writer = Writer()

        writer.writeXmlFile(oglProject=oglProject, fqFileName=self._fileToExport)




from abc import ABC
from abc import abstractmethod

from pyutplugins.plugininterfaces.BasePluginInterface import BasePluginInterface

from pyutplugins.IPluginAdapter import IPluginAdapter


class ToolPluginInterface(BasePluginInterface, ABC):
    """
    This interface defines the methods and properties that Pyut Tool
    pyutplugins must implement.
    """

    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter=pluginAdapter)

        self._menuTitle: str = 'Not Set'

    def executeTool(self):
        """
        This is used by Pyut to invoke the tool.  This should NOT
        be overridden
        TODO: Check for active frame
        """
        if self.setOptions() is True:
            self.doAction()

    @property
    def menuTitle(self) -> str:
        return self._menuTitle

    @abstractmethod
    def setOptions(self) -> bool:
        """
        Prepare for the tool action
        This can be used to query the user for additional plugin options

        Returns: If False, the import should be cancelled.  'True' to proceed
        """
        pass

    @abstractmethod
    def doAction(self):
        """
        Do the tool's action
        """
        pass




from typing import cast

from logging import Logger
from logging import getLogger

from wx.lib.sized_controls import SizedPanel

from pyutplugins.common.ui.DimensionsControl import DimensionsControl
from pyutplugins.common.ui.BaseEditDialog import BaseEditDialog

from pyutplugins.preferences.PluginPreferences import PluginPreferences

from pyutplugins.toolplugins.orthogonal.LayoutAreaSize import LayoutAreaSize


class DlgLayoutSize(BaseEditDialog):

    def __init__(self, parent):

        super().__init__(parent, title='Layout Size')

        self.logger:       Logger          = getLogger(__name__)

        self._preferences:  PluginPreferences = PluginPreferences()

        layoutAreaSize: LayoutAreaSize = self._preferences.orthogonalLayoutSize
        self._layoutWidth:  int = layoutAreaSize.width
        self._layoutHeight: int = layoutAreaSize.height

        self._layoutSizeControl: DimensionsControl = cast(DimensionsControl, None)

        self._layoutSizeControls(parent=self.GetContentsPane())
        self._layoutStandardOkCancelButtonSizer()
        self.Fit()
        self.SetMinSize(self.GetSize())

    @property
    def layoutWidth(self) -> int:
        return self._layoutWidth

    @property
    def layoutHeight(self) -> int:
        return self._layoutHeight

    def _layoutSizeControls(self, parent: SizedPanel):

        self._layoutSizeControl = DimensionsControl(sizedPanel=parent, displayText="Layout Width/Height",
                                                    minValue=480, maxValue=4096,
                                                    valueChangedCallback=self._onSizeChange,
                                                    setControlsSize=True)

        self._layoutSizeControl.dimensions = self._preferences.orthogonalLayoutSize

    def _onSizeChange(self, newValue: LayoutAreaSize):

        self._layoutWidth  = newValue.width
        self._layoutHeight = newValue.height

        self._preferences.orthogonalLayoutSize = newValue




from dataclasses import dataclass

from pyutplugins.plugintypes.ImportDirectoryResponse import ImportDirectoryResponse


@dataclass
class ExportDirectoryResponse(ImportDirectoryResponse):
    """
    Syntactic Sugar
    """




from dataclasses import dataclass


@dataclass
class DTDAttribute:

    __slots__ = ['elementName', 'attributeName', 'attributeType', 'attributeValue', 'valueType']

    def __init__(self):
        pass

    elementName:    str
    attributeName:  str
    attributeType:  str
    attributeValue: str
    valueType:      int





from abc import ABC
from abc import abstractmethod

from dataclasses import dataclass

from pyutplugins.ExternalTypes import CurrentProjectCallback
from pyutplugins.ExternalTypes import OglObjectType
from pyutplugins.ExternalTypes import FrameInformationCallback
from pyutplugins.ExternalTypes import FrameSizeCallback
from pyutplugins.ExternalTypes import PluginProject
from pyutplugins.ExternalTypes import SelectedOglObjectsCallback


@dataclass
class ScreenMetrics:
    screenWidth:  int = 0
    screenHeight: int = 0

    dpiX: int = 0
    dpiY: int = 0


class IPluginAdapter(ABC):
    """
    This the interface specification that allows the pyutplugins to manipulate the Pyut UML Frame
    The Pyut application must implement this and override the appropriate methods and/or
    set appropriate protected variables after call this class
    constructor

    """
    def __init__(self):
        pass

    @property
    @abstractmethod
    def pyutVersion(self) -> str:
        """
        Returns:  The current Pyut version
        """
        pass

    @property
    @abstractmethod
    def screenMetrics(self) -> ScreenMetrics:
        """
        Returns:  appropriate metrics;  wxPython is a helpe
        """
        pass

    @property
    @abstractmethod
    def currentDirectory(self) -> str:
        """
        Returns:  The current directory
        """
        pass

    @abstractmethod
    def getFrameSize(self, callback: FrameSizeCallback):
        pass

    @abstractmethod
    def getFrameInformation(self, callback: FrameInformationCallback):
        pass

    @abstractmethod
    def getSelectedOglObjects(self, callback: SelectedOglObjectsCallback):
        """
        Requests all the selected in the currently displayed frame

        Args:
            callback:  This method is invoked with a list of all the selected OglObjects
        """
        pass

    @abstractmethod
    def refreshFrame(self):
        """
        Refresh the currently displayed frame
        """
        pass

    @abstractmethod
    def selectAllOglObjects(self):
        """
        Select all the Ogl shapes in the currently displayed frame
        """
        pass

    @abstractmethod
    def deselectAllOglObjects(self):
        """
        Deselect all the Ogl shapes in the currently displayed frame
        """
        pass

    @abstractmethod
    def addShape(self, shape: OglObjectType):
        """
        Add an Ogl shape to the currently displayed frame
        Args:
            shape:
        """
        pass

    @abstractmethod
    def loadProject(self, pluginProject: PluginProject):
        """
        Abstract
        This is the preferred way for pyutplugins to projects into Pyut

        Args:
            pluginProject:

        """
        pass

    @abstractmethod
    def requestCurrentProject(self, callback: CurrentProjectCallback):
        """
        Request the current project.   The adapter or its surrogate
        has to convert from a PyutProject to a PluginProject type
        """
        pass

    @abstractmethod
    def indicatePluginModifiedProject(self):
        """
        Plugins always work on the current frame or project
        """
        pass




from pyutplugins.toolplugins.sugiyama.ALayoutLink import ALayoutLink

from miniogl.ControlPoint import ControlPoint


class SugiyamaLink(ALayoutLink):
    """
    SugiyamaLink: link of the Sugiyama graph.

    Instantiated by: ../ToSugiyama.py

    :author: Nicolas Dubois
    :contact: nicdub@gmx.ch
    :version: $Revision: 1.4 $
    """
    def __init__(self, oglObject):
        """
        Constructor.

        @author Nicolas Dubois
        """
        # Call father's initialization
        ALayoutLink.__init__(self, oglObject)
        self.__virtualNodes = []

    def fixControlPoints(self):
        """
        Fix a graphical path with control points.

        @author Nicolas Dubois
        """
        # Clear the actual control points of the link (not the anchor points)
        self.removeAllControlPoints()

        # Current x coordinate of the link
        x = self.getSrcAnchorPos()[0]

        # For all virtual nodes, add control points to pass through
        for vnode in self.__virtualNodes:
            #  ~ print "Virtual node"
            (xvnode, yvnode) = vnode.getPosition()
            # If link goes to up-left
            if x > xvnode:
                # Find the first real node on the right of the virtual node
                neighbor = vnode.getRightNode()
                #
                # Don't like embedded imports, but need to avoid cyclical dependency
                from pyutplugins.toolplugins.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode

                while isinstance(neighbor, VirtualSugiyamaNode) and neighbor is not None:

                    # Try next neighbor
                    neighbor = neighbor.getRightNode()

                # If real node found
                if neighbor is not None:
                    ctrlPoint = ControlPoint(xvnode, neighbor.getPosition()[1] + neighbor.getSize()[1])
                    self.addControlPoint(ctrlPoint)

            else:   # If link goes to up-right
                # Don't like embedded imports, but need to avoid cyclical dependency
                from pyutplugins.toolplugins.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode
                # Find the first real node on the left of the virtual node
                neighbor = vnode.getLeftNode()
                while isinstance(neighbor, VirtualSugiyamaNode) and neighbor is not None:

                    # Try next neighbor
                    neighbor = neighbor.getLeftNode()

                # If real node found
                if neighbor is not None:
                    ctrlPoint = ControlPoint(xvnode, neighbor.getPosition()[1] + neighbor.getSize()[1])
                    self.addControlPoint(ctrlPoint)

            ctrlPoint = ControlPoint(xvnode, yvnode)
            self.addControlPoint(ctrlPoint)

    def addVirtualNode(self, node):
        """
        Add a virtual node.

        A virtual node is inserted in long links which cross a level. If the
        link crosses more than one level, insert virtual nodes, ordered
        from source to destination (son to father - bottom-up).

        @param VirtualSugiyamaNode node : virtual node
        @author Nicolas Dubois
        """
        self.__virtualNodes.append(node)



import sys, os, re, shutil

def main(argv):
    fix("Python3Lexer.g4")
    fix("Python3Parser.g4")

def fix(file_path):
    print("Altering " + file_path)
    if not os.path.exists(file_path):
        print(f"Could not find file: {file_path}")
        sys.exit(1)
    parts = os.path.split(file_path)
    file_name = parts[-1]
    shutil.move(file_path, file_path + ".bak")
    input_file = open(file_path + ".bak",'r')
    output_file = open(file_path, 'w')
    for x in input_file:
        if '!this.' in x:
            x = x.replace('!this.', 'not self.')
        if 'this.' in x:
            x = x.replace('this.', 'self.')
        output_file.write(x)
        output_file.flush()
    print("Writing ...")
    input_file.close()
    output_file.close()

if __name__ == '__main__':
    main(sys.argv)




from logging import Logger
from logging import getLogger

from typing import Tuple

from wx import CENTRE
from wx import OK

from wx import MessageBox

from wx import Yield as wxYield

from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.toolplugins.sugiyama.SugiyamaNode import SugiyamaNode


def cmp(left, right):
    """
        Python 2 replacement

    Args:
        left:
        right:

    Returns:
        -1 if left < right

        0 if left = right

        1 if left > right
    """
    return (left > right) - (left < right)


class SugiyamaGlobals:

    clsLogger: Logger = getLogger(__name__)

    @staticmethod
    def cmpIndex(aTuple: Tuple):
        """
        Internal comparison function for sorting list of parents or children by index.

        Args:
            aTuple:
        Returns:
        """
        sugiyamaNode: SugiyamaNode = aTuple[0]
        l: SugiyamaNode = sugiyamaNode.getLeftNode()
        r: SugiyamaNode = sugiyamaNode.getRightNode()

        if l is None or r is None:
            return 0
        else:
            SugiyamaGlobals.clsLogger.info(f' l.getIndex(): {l.getIndex()}  r.getIndex(): {r.getIndex()}')
            return cmp(l.getIndex(), r.getIndex())

    @staticmethod
    def cmpBarycenter(theNode: SugiyamaNode) -> int:
        """
        Comparison function on barycenter value
        Args:
            theNode

        Returns: The return value from cmp()
        """
        xNode: SugiyamaNode = theNode.getLeftNode()
        yNode: SugiyamaNode = theNode.getRightNode()

        if xNode is None or yNode is None:
            return 0
        else:
            return cmp(xNode.getBarycenter(), yNode.getBarycenter())

    @staticmethod
    def waitKey(pluginAdapter: IPluginAdapter, optionalMessage: str | None):
        # input('Press enter to continue')
        if optionalMessage is None:
            MessageBox('Press Ok to continue', 'Confirm', style=OK | CENTRE)
        else:
            MessageBox(optionalMessage, 'Press Ok to continue', style=OK | CENTRE)
        pluginAdapter.refreshFrame()
        wxYield()



from typing import Tuple

from pyutplugins.toolplugins.sugiyama.SugiyamaNode import SugiyamaNode
from pyutplugins.toolplugins.sugiyama.SugiyamaNode import SugiyamaVEs

from pyutplugins.toolplugins.sugiyama.ALayoutNode import ALayoutNode

from pyutplugins.toolplugins.sugiyama.SugiyamaGlobals import SugiyamaGlobals


class RealSugiyamaNode(SugiyamaNode):
    """
    RealSugiyamaNode: A RealSugiyamaNode object is a node of the Sugiyama
    graph associated to an OglObject of the UML diagram, which can be a
    class or a note.

    For more information, see ../ToSugiyama.py

    Instantiated by: ../ToSugiyama.py

    :author: Nicolas Dubois
    :contact: nicdub@gmx.ch
    :version: $Revision: 1.4 $
    """
    def __init__(self, oglObject):
        """

        Args:
            oglObject: oglObject: diagram class or note
        """
        super().__init__()

        self.__aLayoutNode = ALayoutNode(oglObject)

    def getSize(self) -> Tuple[int, int]:
        """
        Get the size of the node.

        Returns: (int, int) : tuple (width, height)
        """
        return self.__aLayoutNode.getSize()

    def setPosition(self, xCoord: int, yCoord: int):
        """
        Set node position.

        Args:
            xCoord:  x position in absolute coordinates
            yCoord:  y position in absolute coordinates
        """
        self.__aLayoutNode.setPosition(xCoord, yCoord)

    def getPosition(self) -> Tuple[int, int]:
        """
        Get node position.

        Returns: (int, int) : tuple (x, y) in absolute coordinates
        """
        return self.__aLayoutNode.getPosition()

    def getName(self) -> str:
        """
        Get the name of the OglObject.

        @return str : name of OglObject
        @author Nicolas Dubois
        """
        return self.__aLayoutNode.getName()

    def fixAnchorPos(self):
        """
        Fix the positions of the anchor points.

        The anchor points are placed according to parent and child positions.
        Before calling this method, be sure you have set the indices of all
        parent and children (see setIndex).

        """
        # Get position and size of node
        (width, height) = self.getSize()
        (x, y) = self.getPosition()

        # Fix all children anchors position
        # Sort child list to eliminate crossing
        children: SugiyamaVEs = self.getChildren()
        children.sort(key=SugiyamaGlobals.cmpIndex)

        nChildren: int = len(children)
        # For all children
        for i in range(nChildren):
            (child, link) = children[i]
            # Fix anchors coordinates
            link.setDestAnchorPos(x + width * (i + 1) // (nChildren + 1), y + height)

        # Parent anchors position
        # Sort parents list to eliminate crossing
        parents: SugiyamaVEs = self.getParents()
        parents.sort(key=SugiyamaGlobals.cmpIndex)
        nParents: int = len(parents)
        # For all parents
        for i in range(nParents):
            (parent, link) = parents[i]
            # Fix anchors coordinates
            link.setSrcAnchorPos(x + width * (i + 1) // (nParents + 1), y)

    def __repr__(self) -> str:
        return f'RealSugiyamaNode name: {self.getName()} level: {self.getLevel()}'




from typing import Callable
from typing import Dict
from typing import List
from typing import NewType
from typing import Tuple
from typing import Union
from typing import cast

from logging import Logger
from logging import getLogger

from os import sep as osSep

from antlr4 import CommonTokenStream
from antlr4 import FileStream

from ogl.OglLink import OglLink
from ogl.OglClass import OglClass

from pyutmodel.PyutLinkType import PyutLinkType
from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutField import PyutField
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutMethod import PyutParameters
from pyutmodel.PyutMethod import SourceCode
from pyutmodel.PyutParameter import PyutParameter
from pyutmodel.PyutType import PyutType
from pyutmodel.PyutVisibilityEnum import PyutVisibilityEnum

from pyutplugins.common.LinkMakerMixin import LinkMakerMixin
from pyutplugins.ExternalTypes import OglClasses
from pyutplugins.ExternalTypes import OglLinks


from pyutplugins.ioplugins.python.PythonParseException import PythonParseException
from pyutplugins.ioplugins.python.PyutPythonVisitor import ChildName
from pyutplugins.ioplugins.python.PyutPythonVisitor import Children
from pyutplugins.ioplugins.python.PyutPythonVisitor import ClassName
from pyutplugins.ioplugins.python.PyutPythonVisitor import ClassNames
from pyutplugins.ioplugins.python.PyutPythonVisitor import DataClassProperties
from pyutplugins.ioplugins.python.PyutPythonVisitor import MethodName
from pyutplugins.ioplugins.python.PyutPythonVisitor import MethodNames
from pyutplugins.ioplugins.python.PyutPythonVisitor import MultiParameterNames
from pyutplugins.ioplugins.python.PyutPythonVisitor import Parameters
from pyutplugins.ioplugins.python.PyutPythonVisitor import ParentName
from pyutplugins.ioplugins.python.PyutPythonVisitor import Parents
from pyutplugins.ioplugins.python.PyutPythonVisitor import PyutPythonVisitor
from pyutplugins.ioplugins.python.pyantlrparser.Python3Lexer import Python3Lexer
from pyutplugins.ioplugins.python.pyantlrparser.Python3Parser import Python3Parser

PyutClassName  = NewType('PyutClassName', str)
PyutClasses    = NewType('PyutClasses', Dict[PyutClassName, PyutClass])
OglClassesDict = NewType('OglClassesDict', Dict[Union[PyutClassName, ParentName, ChildName, ClassName], OglClass])


class ReverseEngineerPython2(LinkMakerMixin):

    PYTHON_ASSIGNMENT:     str = '='
    PYTHON_TYPE_DELIMITER: str = ':'
    PYTHON_EOL_COMMENT:    str = '#'

    def __init__(self):

        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._pyutClasses:    PyutClasses     = PyutClasses({})
        self._oglClassesDict: OglClassesDict  = OglClassesDict({})
        self._oglClasses:     OglClasses      = cast(OglClasses, None)
        self._oglLinks:       OglLinks        = OglLinks([])

        self.visitor: PyutPythonVisitor = cast(PyutPythonVisitor, None)

    def reversePython(self,  directoryName: str, files: List[str], progressCallback: Callable):
        """
        Reverse engineering Python files to OglClass's

        Args:
            directoryName:  The directory name where the selected files reside
            files:          A list of files to parse
            progressCallback: The method to call to report progress
        """
        currentFileCount: int = 0

        onGoingParents: Parents = Parents({})
        for fileName in files:

            try:
                fqFileName: str = f'{directoryName}{osSep}{fileName}'
                self.logger.info(f'Processing file: {fqFileName}')

                progressCallback(currentFileCount, f'Processing: {fileName}')

                fileStream: FileStream   = FileStream(fqFileName)
                lexer:      Python3Lexer = Python3Lexer(fileStream)

                stream: CommonTokenStream = CommonTokenStream(lexer)
                parser: Python3Parser     = Python3Parser(stream)

                tree: Python3Parser.File_inputContext = parser.file_input()
                if parser.getNumberOfSyntaxErrors() != 0:
                    eMsg: str = f"File {fileName} contains {parser.getNumberOfSyntaxErrors()} syntax errors"
                    self.logger.error(eMsg)
                    raise PythonParseException(eMsg)

                self.visitor = PyutPythonVisitor()
                self.visitor.parents = onGoingParents
                self.visitor.visit(tree)
                self._generatePyutClasses()

                onGoingParents = self.visitor.parents
                currentFileCount += 1
            except (ValueError, Exception) as e:
                self.logger.error(e)
                raise PythonParseException(e)
        self._generateOglClasses()
        self._generateInheritanceLinks()

    @property
    def oglClasses(self) -> OglClasses:
        if self._oglClasses is None:
            self._oglClasses = list(self._oglClassesDict.values())

        return self._oglClasses

    def oglLinks(self) -> OglLinks:
        return self._oglLinks

    def _generateOglClasses(self):

        for pyutClassName in self._pyutClasses:
            try:
                pyutClass: PyutClass = self._pyutClasses[pyutClassName]
                oglClass:  OglClass  = OglClass(pyutClass)

                self._oglClassesDict[pyutClassName] = oglClass
            except (ValueError, Exception) as e:
                self.logger.error(f"Error while creating class {pyutClassName},  {e}")

    def _generatePyutClasses(self):

        for className in self._classNames():
            pyutClass: PyutClass = PyutClass(name=className)

            pyutClass = self._addFields(pyutClass)

            for methodName in self._methodNames(className):
                pyutMethod: PyutMethod = PyutMethod(name=methodName)
                if methodName[0:2] == "__":
                    pyutMethod.setVisibility(PyutVisibilityEnum.PRIVATE)
                elif methodName[0] == "_":
                    pyutMethod.setVisibility(PyutVisibilityEnum.PROTECTED)
                else:
                    pyutMethod.setVisibility(PyutVisibilityEnum.PUBLIC)
                pyutMethod = self._addParameters(pyutMethod)
                pyutMethod.sourceCode = SourceCode(self.visitor.methodCode[methodName])

                pyutClass.addMethod(pyutMethod)

            setterProperties: Parameters = self.visitor.setterProperties
            getterProperties: Parameters = self.visitor.getterProperties

            pyutClass = self._generatePropertiesAsMethods(pyutClass, getterProperties, setterProperties)

            if className in self.visitor.dataClassNames:
                self._createDataClassPropertiesAsFields(pyutClass, self.visitor.dataClassProperties)

            self._pyutClasses[PyutClassName(className)] = pyutClass
        self.logger.info(f'Generated {len(self._pyutClasses)} classes')

    def _generatePropertiesAsMethods(self, pyutClass: PyutClass, getterProperties, setterProperties) -> PyutClass:

        for propName in self.visitor.propertyNames:
            # Might be a read-only property
            try:
                setterParams: List[str] = setterProperties[propName]
            except KeyError:
                setterParams = []

            getterParams: List[str] = getterProperties[propName]
            self.logger.info(f'Processing - {propName=} {setterParams=} {getterParams=}')
            setter, getter = self._createProperties(propName=propName, setterParams=setterParams)
            if setter is not None:
                setter.isProperty = True
                pyutClass.addMethod(setter)
            getter.isProperty = True
            pyutClass.addMethod(getter)

        return pyutClass

    def _createProperties(self, propName: str, setterParams: List[str]) -> Tuple[PyutMethod, PyutMethod]:

        getter: PyutMethod = PyutMethod(name=propName, visibility=PyutVisibilityEnum.PUBLIC)
        if len(setterParams) == 0:
            setter: PyutMethod = cast(PyutMethod, None)
        else:
            setter = PyutMethod(name=propName, visibility=PyutVisibilityEnum.PUBLIC)

        if setter is not None:
            nameType: str = setterParams[0]
            potentialNameType: List[str] = nameType.split(':')

            if len(potentialNameType) == 2:

                param: PyutParameter = PyutParameter(name=potentialNameType[0], parameterType=PyutType(value=potentialNameType[1]))
                setter.addParameter(param)
                getter.returnType = PyutType(value=potentialNameType[1])
            else:
                param = PyutParameter(name=potentialNameType[0])
                setter.addParameter(param)

        return setter, getter

    def _createDataClassPropertiesAsFields(self, pyutClass: PyutClass, dataClassProperties: DataClassProperties) -> PyutClass:
        """

        Args:
            pyutClass:  The PyutClass to update
            dataClassProperties:    The dataclass properties to parse

        Returns:
            Updated PyutClass with new fields
        """
        className: str = pyutClass.name
        for dPropertyTuple in dataClassProperties:
            if dPropertyTuple[0] == className:

                pyutField: PyutField = self._parseFieldToPyut(fieldData=dPropertyTuple[1])
                pyutClass.addField(pyutField)

        return pyutClass

    def _addParameters(self, pyutMethod: PyutMethod) -> PyutMethod:

        methodName: MethodName = MethodName(pyutMethod.name)
        if methodName in self.visitor.parameters:
            multiParameterNames: MultiParameterNames = self.visitor.parameters[methodName]

            pyutParameters: PyutParameters = self._generateParameters(multiParameterNames)
            pyutMethod.parameters = pyutParameters

        return pyutMethod

    def _addFields(self, pyutClass: PyutClass) -> PyutClass:
        """
        Can look like this:

           fieldData: x:int=0
           fieldData = 0

        Args:
            pyutClass:  Where to add the fields

        Returns:  The updated input class

        """
        for fieldData in self.visitor.fields:
            self.logger.debug(f'fieldData: {fieldData}')
            pyutField: PyutField = self._parseFieldToPyut(fieldData)
            pyutClass.addField(pyutField)

        return pyutClass

    def _generateInheritanceLinks(self):

        parents: Parents = self.visitor.parents

        for parentName in parents.keys():
            children: Children = parents[parentName]
            for childName in children:

                try:
                    parentOglClass: OglClass = self._oglClassesDict[parentName]
                    childOglClass:  OglClass = self._oglClassesDict[childName]
                    oglLink: OglLink = self.createLink(src=childOglClass, dst=parentOglClass, linkType=PyutLinkType.INHERITANCE)
                    self._oglLinks.append(oglLink)
                except KeyError as ke:        # Probably there is no parent we are tracking
                    self.logger.warning(f'Apparently we are not tracking this parent:  {ke}')
                    continue

    def _methodNames(self, className: ClassName) -> MethodNames:

        methodNames: MethodNames = MethodNames([])
        try:
            methodNames = self.visitor.classMethods[className]
        except KeyError:
            self.logger.warning(f'{className} has no methods')

        return methodNames

    def _classNames(self) -> ClassNames:
        return self.visitor.classNames

    def _parseFieldToPyut(self, fieldData: str) -> PyutField:

        self.logger.debug(f'fieldData: {fieldData}')

        if ReverseEngineerPython2.PYTHON_TYPE_DELIMITER in fieldData and ReverseEngineerPython2.PYTHON_ASSIGNMENT in fieldData:
            pyutField: PyutField = self.__complexParseFieldToPyut(fieldData)
        else:
            pyutField = self.__simpleParseFieldToPyut(fieldData)

        return pyutField

    def _generateParameters(self, multiParameterNames: str) -> PyutParameters:
        """
        Handles the following 4 cases:
        Simple:                       param
        Typed:                        param: float
        SimpleDefaultValue:           param=0.0
        ComplexTypedAndDefaultValue: param: float = 0.0
        Args:
            multiParameterNames:

        Returns:  A list of PyutParam objects
        """
        parameterNameList: List[str] = multiParameterNames.split(',')
        pyutParams:        PyutParameters = PyutParameters([])
        for parameterStr in parameterNameList:
            if ':' in parameterStr and '=' in parameterStr:
                pyutParam: PyutParameter = self.__complexTypedAndDefaultValue(parameterStr)
            elif '=' in parameterStr:
                pyutParam = self._simpleDefaultValue(parameterStr)
            elif ':' in parameterStr:
                pyutParam = self._typedParameter(parameterStr)
            else:
                pyutParam = PyutParameter(parameterStr)
            pyutParams.append(pyutParam)

        return pyutParams

    def __complexTypedAndDefaultValue(self, complexParam: str) -> PyutParameter:

        paramNameType:  List[str] = complexParam.split(':')
        paramName:      str = paramNameType[0]
        paramTypeValue: List[str] = paramNameType[1].split('=')
        paramType:      str = paramTypeValue[0]
        paramValue:     str = paramTypeValue[1]

        pyutType: PyutType = PyutType(paramType)
        return PyutParameter(name=paramName, parameterType=pyutType, defaultValue=paramValue)

    def _simpleDefaultValue(self, simpleDefaultValueParam: str) -> PyutParameter:

        pyutParamAndValue: List[str] = simpleDefaultValueParam.split('=')
        paramName:  str = pyutParamAndValue[0]
        paramValue: str = pyutParamAndValue[1]

        pyutParam: PyutParameter = PyutParameter(name=paramName, defaultValue=paramValue)

        return pyutParam

    def _typedParameter(self, typedParam: str) -> PyutParameter:
        pyutParamAndType: List[str] = typedParam.split(':')
        paramName:        str = pyutParamAndType[0]
        paramType:        str = pyutParamAndType[1]

        pyutParam: PyutParameter = PyutParameter(name=paramName, parameterType=PyutType(value=paramType))
        return pyutParam

    def __simpleParseFieldToPyut(self, fieldData: str) -> PyutField:

        pyutField: PyutField = PyutField()

        noCommentFieldData: str = self.__stripEndOfLineComment(fieldData)
        fieldAndValue: List[str] = noCommentFieldData.split(ReverseEngineerPython2.PYTHON_ASSIGNMENT)

        if len(fieldAndValue) == 2:
            pyutField.name         = fieldAndValue[0].strip()
            pyutField.defaultValue = fieldAndValue[1].strip()
        else:   # might just be a declaration
            pyutField = self.__declarationOnlyParseToPyut(fieldData=fieldAndValue[0])
        return pyutField

    def __complexParseFieldToPyut(self, fieldData: str) -> PyutField:
        """
          Can look like this:

           fieldData: x:int=0

        Args:
            fieldData:

        Returns:
        """
        noCommentFieldData: str = self.__stripEndOfLineComment(fieldData)

        fieldAndType: List[str] = noCommentFieldData.split(ReverseEngineerPython2.PYTHON_TYPE_DELIMITER)
        fieldName:    str       = fieldAndType[0]

        vis: PyutVisibilityEnum = self.__determineFieldVisibility(fieldName)

        fieldName = self.__appropriatelyCleanupName(vis=vis, fieldName=fieldName)

        pyutField: PyutField = PyutField(name=fieldName, visibility=vis)

        if len(fieldAndType) > 1:
            typeAndDefaultValue: List[str] = fieldAndType[1].split(ReverseEngineerPython2.PYTHON_ASSIGNMENT)

            pyutType: PyutType = PyutType(value=typeAndDefaultValue[0].strip())
            pyutField.type = pyutType
            if len(typeAndDefaultValue) > 1:
                pyutField.defaultValue = typeAndDefaultValue[1].strip()

        return pyutField

    def __declarationOnlyParseToPyut(self, fieldData: str) -> PyutField:

        self.logger.info(f'{fieldData=}')
        fieldAndType: List[str] = fieldData.split(ReverseEngineerPython2.PYTHON_TYPE_DELIMITER)

        pyutField: PyutField = PyutField(name=fieldAndType[0])

        #
        # Might be something complex expression as a default value we can't handle it
        #
        if len(fieldAndType) > 1:
            pyutField.type       = PyutType(value=fieldAndType[1])
        pyutField.visibility = PyutVisibilityEnum.PUBLIC

        return pyutField

    def __determineFieldVisibility(self, name: str) -> PyutVisibilityEnum:

        vis: PyutVisibilityEnum = PyutVisibilityEnum.PUBLIC
        if len(name) > 1:
            if name[-2:] != "__":
                if name[0:2] == "__":
                    vis = PyutVisibilityEnum.PRIVATE
                elif name[0] == "_":
                    vis = PyutVisibilityEnum.PROTECTED
        return vis

    def __appropriatelyCleanupName(self, vis: PyutVisibilityEnum, fieldName: str) -> str:

        if vis == PyutVisibilityEnum.PUBLIC:
            return fieldName
        elif vis == PyutVisibilityEnum.PRIVATE:
            fieldName = fieldName[2:]
            return fieldName
        else:
            fieldName = fieldName[1:]
            return fieldName

    def __stripEndOfLineComment(self, fieldData: str) -> str:

        fieldAndComment: List[str] = fieldData.split(ReverseEngineerPython2.PYTHON_EOL_COMMENT)

        return fieldAndComment[0]




from typing import List
from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from enum import Enum

from os import linesep as eol

from datetime import datetime

from pathlib import Path

from pyutmodel.PyutField import PyutField
from pyutmodel.PyutField import PyutFields
from pyutmodel.PyutLink import PyutLink
from pyutmodel.PyutType import PyutType
from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutParameter import PyutParameter
from pyutmodel.PyutStereotype import PyutStereotype
from pyutmodel.PyutVisibilityEnum import PyutVisibilityEnum

from ogl.OglLink import OglLink
from ogl.OglClass import OglClass
from ogl.OglInterface2 import OglInterface2
from ogl.OglAggregation import OglAggregation
from ogl.OglInheritance import OglInheritance
from ogl.OglComposition import OglComposition
from ogl.OglAssociation import OglAssociation
from ogl.OglInterface import OglInterface

from pyutplugins.ExternalTypes import OglObjects
from pyutplugins.ioplugins.mermaid.MermaidDirection import MermaidDirection
from pyutplugins.preferences.PluginPreferences import PluginPreferences


indent1: str = '    '
indent2: str = f'{indent1}    '
indent3: str = f'{indent2}    '
indent4: str = f'{indent3}    '
indent5: str = f'{indent4}    '


class MermaidArrow(Enum):
    INHERITANCE_ARROW = '<|--'  # Points to parent class
    AGGREGATION_LINK  = 'o--'   #
    COMPOSITION_LINK  = '*--'   #
    INTERFACE_LINK    = '..|>'  #
    ASSOCIATION_LINK  = '--'


class MermaidWriter:
    VERSION: str = '0.5'

    def __init__(self, fqFileName: Path, writeCredits: bool = True):
        """

        Args:
            fqFileName:     Fully qualified file name to write to
            writeCredits:   We will always write credits.  Unit tests turn this off
            because of the time stamps
        """
        self.logger: Logger = getLogger(__name__)

        self._fqFileName: Path = fqFileName

        self._preferences: PluginPreferences = PluginPreferences()

        self._writeCredits(writeCredits=writeCredits)

    def translate(self, oglObjects: OglObjects):

        mermaidString: str              = f'```mermaid{eol}'
        direction:     MermaidDirection = self._preferences.mermaidLayoutDirection
        mermaidString = f'{mermaidString}classDiagram{eol}{indent1}{direction.diagramValue}{eol}'

        linksStanza: str = self._generateLinksStanza(oglObjects=oglObjects)

        mermaidString = f'{mermaidString}{linksStanza}'

        for oglObject in oglObjects:
            match oglObject:
                case OglClass():
                    oglClass:  OglClass  = cast(OglClass, oglObject)
                    generatedString: str = self._generateClassStanza(oglClass)
                    mermaidString += f'{generatedString}'
                case OglLink():
                    pass
                case _:
                    self.logger.warning(f'Unsupported Ogl element: {oglObject}')

        mermaidString += f'```{eol}'
        with self._fqFileName.open(mode='a') as fd:
            fd.write(f'{mermaidString}')

    def _generateClassStanza(self, oglObject: OglClass) -> str:
        """

        Args:
            oglObject:  A Pyut Class Definition

        Returns:  A Mermaid class definition
        """
        oglClass:  OglClass  = cast(OglClass, oglObject)
        pyutClass: PyutClass = oglClass.pyutObject

        fields:         PyutFields      = pyutClass.fields
        fieldsRefrain:  str              = self._generateFieldsRefrain(fields)
        methods:        List[PyutMethod] = pyutClass.methods
        methodsStr:     str              = self._generateMethods(methods)

        if pyutClass.stereotype == PyutStereotype.NO_STEREOTYPE:
            stereotypeRefrain: str = ''
        else:
            stereotypeRefrain = f'{indent2}<<{pyutClass.stereotype.value}>>{eol}'

        generatedString: str = (
            f'{indent1}class {pyutClass.name} {{ {eol}'
            f'{fieldsRefrain}'
            f'{stereotypeRefrain}'
            f'{methodsStr}'
            f'{indent1}}}{eol}'
        )
        return generatedString

    def _generateLinksStanza(self, oglObjects: OglObjects) -> str:
        """
        Make a pass through creating links
        Args:
            oglObjects:

        Returns:

        """
        linksStanza: str = f'{indent1}%% Links follow{eol}'

        for oglObject in oglObjects:
            linkRefrain: str = ''
            match oglObject:
                case OglInheritance():
                    linkRefrain = self._getInheritanceLinkRefrain(oglLink=oglObject)
                case OglAggregation():
                    oglAggregation: OglAggregation = cast(OglAggregation, oglObject)
                    linkRefrain = self._getAssociationLinkRefrain(oglAssociation=oglAggregation, arrowType=MermaidArrow.AGGREGATION_LINK)
                case OglComposition():
                    oglComposition: OglComposition = cast(OglComposition, oglObject)
                    linkRefrain = self._getAssociationLinkRefrain(oglAssociation=oglComposition, arrowType=MermaidArrow.COMPOSITION_LINK)
                case OglInterface():
                    oglInterface: OglInterface = cast(OglInterface, oglObject)
                    linkRefrain = self._getRealizationLinkRefrain(oglInterface)
                case OglAssociation():      # Most general needs to be last
                    oglAssociation: OglAssociation = cast(OglAssociation, oglObject)
                    linkRefrain = self._getAssociationLinkRefrain(oglAssociation=oglAssociation, arrowType=MermaidArrow.ASSOCIATION_LINK)
                case OglInterface2():
                    pass
                case _:
                    pass        # Ignore non links
            linksStanza += linkRefrain

        return linksStanza

    def _getAssociationLinkRefrain(self, oglAssociation: OglAssociation, arrowType: MermaidArrow) -> str:
        """
        For Composition and Aggregation the diamond is on the source side
        Produces refrains of the form:

        For composition:
        Person "1" *-- "1" Heart

        Where Person is the 'composer' and Heart is the 'composed'

        For aggregation

        Author "1.*" o-- "0.*" Book

        Where Author aggregates Books

        Args:
            oglAssociation:
            arrowType:

        Returns:
        """
        pyutLink: PyutLink = oglAssociation.pyutObject
        sourceName:      str = pyutLink.getSource().name
        destinationName: str = pyutLink.getDestination().name
        self.logger.info(f'{oglAssociation} {sourceName=} {destinationName=}')

        sourceCardinality, destinationCardinality = self._getCardinalityStrings(pyutLink)
        linkRefrain = (
            f'{indent1}{sourceName} {sourceCardinality} {arrowType.value} {destinationCardinality} {destinationName} : {pyutLink.name}{eol}'
        )
        return linkRefrain

    def _getInheritanceLinkRefrain(self, oglLink: OglLink) -> str:
        """
        Produces strings of the form:

        Animal<|--Duck

        Where Animal is the parent class and Duck is the subclass

        Args:
            oglLink:  The inheritance link

        Returns:  Mermaid string for inheritance
        """
        pyutLink: PyutLink = oglLink.pyutObject

        subClassName:  str = pyutLink.getSource().name
        baseClassName: str = pyutLink.getDestination().name
        self.logger.info(f'{subClassName=} {pyutLink.linkType=} {baseClassName=}')

        linkRefrain: str  = (
            f'{indent1}{baseClassName}{MermaidArrow.INHERITANCE_ARROW.value}{subClassName}{eol}'
        )

        return linkRefrain

    def _getRealizationLinkRefrain(self, oglInterface: OglInterface) -> str:
        pyutLink: PyutLink = oglInterface.pyutObject
        interfaceName:   str = pyutLink.getSource().name
        implementorName: str = pyutLink.getDestination().name
        linkRefrain: str = (
            f'{indent1}{interfaceName} {MermaidArrow.INTERFACE_LINK.value} {implementorName}{eol}'
        )

        return linkRefrain

    def _getCardinalityStrings(self, pyutLink) -> Tuple[str, str]:
        """

        Args:
            pyutLink:

        Returns: A tuple of source, destination cardinality strings

        """
        if pyutLink.sourceCardinality == '':
            sourceCardinality: str = ''
        else:
            sourceCardinality = f'"{pyutLink.sourceCardinality}"'

        if pyutLink.destinationCardinality == '':
            destinationCardinality: str = ''
        else:
            destinationCardinality = f'"{pyutLink.destinationCardinality}"'

        return sourceCardinality, destinationCardinality

    def _generateFieldsRefrain(self, fields: PyutFields) -> str:
        fieldsRefrain: str = ''
        for field in fields:
            pyutField: PyutField = cast(PyutField, field)
            visibility: PyutVisibilityEnum = pyutField.visibility
            retType:    PyutType = pyutField.type

            fieldString: str = (
                f'{indent2}{visibility.value}{retType.value} {pyutField.name}{eol}'
            )
            fieldsRefrain += fieldString
        return fieldsRefrain

    def _generateMethods(self, methods: List[PyutMethod]) -> str:

        retString: str = f''

        for method in methods:
            pyutMethod: PyutMethod = cast(PyutMethod, method)
            visibility: PyutVisibilityEnum = pyutMethod.visibility
            retType:    PyutType = pyutMethod.returnType

            parameterString: str = self._generateParameters(pyutMethod)
            methodString: str = (
                f'{indent2}{visibility.value}{retType.value} {pyutMethod.name}'
                f'('
                f'{parameterString}'
                f'){eol}'
            )
            retString += methodString
        return retString

    def _generateParameters(self, pyutMethod: PyutMethod) -> str:
        retString: str = ''

        for parameter in pyutMethod.parameters:
            pyutParameter: PyutParameter = cast(PyutParameter, parameter)
            parameterString: str = (
                f'{pyutParameter.name}, '
            )
            retString += parameterString
        return retString.strip(', ')

    def _writeCredits(self, writeCredits: bool):

        # ensure it is empty
        with self._fqFileName.open(mode='w') as fd:
            if writeCredits is True:

                fd.write(f'---{eol}')
                fd.write(f'Generated by Mermaid Writer Version: {MermaidWriter.VERSION}{eol}')

                now:      datetime = datetime.now()  # current date and time
                dateTime: str      = now.strftime('%d %b %Y, %H:%M:%S')

                fd.write(f'{dateTime}{eol}')
                fd.write(f'---{eol}')
            else:
                fd.write('')




from typing import cast

from logging import Logger
from logging import getLogger

from wx import MessageBox
# noinspection PyProtectedMember
from wx._core import BitmapType

from wx import Bitmap
from wx import ClientDC
from wx import Image
from wx import MemoryDC
from wx import NullBitmap
from wx import OK

from pyutplugins.plugintypes.PluginDataTypes import FormatName
from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import PluginName

from pyutplugins.ExternalTypes import FrameInformation
from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.IPluginAdapter import IPluginAdapter
from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface

from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat

from pyutplugins.ioplugins.wximage.DlgWxImageOptions import DlgWxImageOptions
from pyutplugins.ioplugins.wximage.WxImageFormat import WxImageFormat

FORMAT_NAME:        FormatName        = FormatName('Wx Image')
PLUGIN_EXTENSION:   PluginExtension   = PluginExtension('png')
PLUGIN_DESCRIPTION: PluginDescription = PluginDescription('png, bmp, gif, or jpg')


class IOWxImage(IOPluginInterface):

    def __init__(self, pluginAdapter: IPluginAdapter):
        """

        Args:
            pluginAdapter:   A class that implements IMediator
        """
        super().__init__(pluginAdapter=pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._name    = PluginName('Wx Image')
        self._author  = 'Humberto A. Sanchez II'
        self._version = '0.9c'

        self._inputFormat  = cast(InputFormat, None)
        self._outputFormat = OutputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)

        self._autoSelectAll = True     # we are taking a picture of the entire diagram

        self._imageFormat:    WxImageFormat = cast(WxImageFormat, None)
        self._outputFileName: str           = cast(str, None)

    def setImportOptions(self) -> bool:
        return False

    def setExportOptions(self) -> bool:
        """
        Popup the options dialog

        Returns:
            if False, the export will be cancelled.
        """
        with DlgWxImageOptions(None) as dlg:
            if dlg.ShowModal() == OK:
                self.logger.warning(f'{dlg.imageFormat=} {dlg.outputFileName=}')
                self._imageFormat    = dlg.imageFormat
                self._outputFileName = dlg.outputFileName

            else:
                self.logger.warning(f'Cancelled')
                return False

        return True

    def read(self) -> bool:
        return False

    def write(self, oglObjects: OglObjects):
        """
        Write data

        Args:
            oglObjects:     list of exported objects
        """
        pluginAdapter:    IPluginAdapter   = self._pluginAdapter
        frameInformation: FrameInformation = self._frameInformation
        pluginAdapter.deselectAllOglObjects()

        imageType: BitmapType     = WxImageFormat.toWxBitMapType(self._imageFormat)

        context:   ClientDC       = frameInformation.clientDC
        memory:    MemoryDC       = MemoryDC()

        x: int = frameInformation.frameSize.width
        y: int = frameInformation.frameSize.height
        emptyBitmap: Bitmap = Bitmap(x, y, -1)

        memory.SelectObject(emptyBitmap)
        memory.Blit(source=context, xsrc=0, height=y, xdest=0, ydest=0, ysrc=0, width=x)
        memory.SelectObject(NullBitmap)

        img:       Image = emptyBitmap.ConvertToImage()
        extension: str   = self._imageFormat.__str__()

        filename: str   = f'{self._outputFileName}.{extension}'
        status:   bool  = img.SaveFile(filename, imageType)
        if status is False:
            msg: str = f'Error on image write to {filename}'
            self.logger.error(msg)
            MessageBox(message=msg, caption='Error', style=OK)




from typing import cast

from logging import Logger
from logging import getLogger

from pyutplugins.plugintypes.PluginDataTypes import PluginName
from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import FormatName

from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat
from pyutplugins.plugintypes.SingleFileRequestResponse import SingleFileRequestResponse

from pyutplugins.ioplugins.gml.GMLExporter import GMLExporter

from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.IPluginAdapter import IPluginAdapter
from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface

FORMAT_NAME:        FormatName = FormatName('GML')
PLUGIN_EXTENSION:   PluginExtension = PluginExtension('gml')
PLUGIN_DESCRIPTION: PluginDescription = PluginDescription('Graph Modeling Language - Portable Format for Graphs')


class IOGML(IOPluginInterface):
    """
    """
    def __init__(self, pluginAdapter: IPluginAdapter):
        """

        Args:
            pluginAdapter:   A class that implements IPluginAdapter
        """
        super().__init__(pluginAdapter=pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._name    = PluginName('Output GML')
        self._author  = "Humberto A. Sanchez II"
        self._version = GMLExporter.VERSION

        self._exportResponse: SingleFileRequestResponse = cast(SingleFileRequestResponse, None)

        self._inputFormat  = cast(InputFormat, None)
        self._outputFormat = OutputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)

        self._autoSelectAll = False      # Temp until we have plugin preferences

    def setImportOptions(self) -> bool:
        """
        Prepare the import.
        This can be used to ask some questions to the user.

        Returns:
            if False, the import will be cancelled.
        """
        return False

    def setExportOptions(self) -> bool:
        """
        Prepare the export.

        Returns:
            if False, the export is cancelled.
        """
        self._exportResponse = self.askForFileToExport()

        if self._exportResponse.cancelled is True:
            return False
        else:
            return True

    def read(self) -> bool:
        """
        Read data from filename.
        """
        return False

    def write(self, oglObjects: OglObjects):
        """
        Write data

        Args:
            oglObjects:     list of exported objects
        """
        gmlExporter: GMLExporter = GMLExporter()

        gmlExporter.translate(umlObjects=oglObjects)

        gmlExporter.write(self._exportResponse.fileName)




from typing import List
from typing import Union

from logging import Logger
from logging import getLogger

from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.plugintypes.PluginDataTypes import PluginName

from pyutplugins.plugininterfaces.ToolPluginInterface import ToolPluginInterface

from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.toolplugins.sugiyama.RealSugiyamaNode import RealSugiyamaNode
from pyutplugins.toolplugins.sugiyama.Sugiyama import Sugiyama
from pyutplugins.toolplugins.sugiyama.SugiyamaLink import SugiyamaLink
from pyutplugins.toolplugins.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode


class ToolSugiyama(ToolPluginInterface):
    """
    ToSugiyama : Automatic layout algorithm based on Sugiyama levels.
    """
    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._name      = PluginName('Sugiyama Automatic Layout')
        self._author    = 'Nicolas Dubois <nicdub@gmx.ch>'
        self._version   = '1.1'

        self._menuTitle = 'Sugiyama Automatic Layout'

        #
        # TODO Move to separate class
        #
        # Sugiyama nodes and links
        self.__realSugiyamaNodesList: List[RealSugiyamaNode] = []   # List of all RealSugiyamaNode
        self.__sugiyamaLinksList:     List[SugiyamaLink]     = []   # List of all SugiyamaLink

        #  Hierarchy graph
        #  List of Real and Virtual Sugiyama nodes that take part in hierarchy
        self.__hierarchyGraphNodesList:    List[Union[RealSugiyamaNode, VirtualSugiyamaNode]] = []
        #  List of Sugiyama nodes that are not in hierarchy
        self.__nonHierarchyGraphNodesList: List[VirtualSugiyamaNode] = []
        self.__nonHierarchyGraphLinksList: List[SugiyamaLink]        = []

        #  All nodes of the hierarchy are assigned to a level.
        #  A level is a list of nodes (real or virtual).
        self.__levels: List = []  # List of levels

    def setOptions(self) -> bool:
        """
        Prepare for the tool action.
        This can be used to ask some questions to the user.

        Returns: If False, the import should be cancelled.  'True' to proceed
        """
        return True

    def doAction(self):
        self._pluginAdapter.getSelectedOglObjects(callback=self._doAction)

    def _doAction(self, selectedOglObjects: OglObjects):

        selectedObjects: OglObjects = selectedOglObjects

        self.logger.info(f'Begin Sugiyama algorithm')

        sugiyama: Sugiyama = Sugiyama(pluginAdapter=self._pluginAdapter)
        sugiyama.createInterfaceOglALayout(oglObjects=selectedObjects)
        sugiyama.levelFind()
        sugiyama.addVirtualNodes()
        sugiyama.barycenter()

        # noinspection PyProtectedMember
        self.logger.info(f'Number of hierarchical intersections: {sugiyama._getNbIntersectAll()}')

        sugiyama.addNonHierarchicalNodes()
        sugiyama.fixPositions()

        self._pluginAdapter.indicatePluginModifiedProject()
        self._pluginAdapter.refreshFrame()

        self.logger.info('End Sugiyama algorithm')




from typing import cast

from abc import ABC
from abc import abstractmethod

from pyutplugins.plugininterfaces.BasePluginInterface import BasePluginInterface
from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.ExternalTypes import FrameInformation
from pyutplugins.ExternalTypes import OglObjects


class IOPluginInterface(BasePluginInterface, ABC):
    """
    Abstract class for input/output plug-ins.

    If you want to do a new plugin, you must inherit from this class and
    implement the abstract methods.

    The plugin may require user interaction for plugin parameters.  Implement
    these methods:

        `setImportOptions`
        `setExportOptions`

    The import/export work is done in:

        `read(self, oglObjects, umlFrame)`
        `write(self, oglObjects)`

    Pyut invokes the plugin, by instantiating it, and calling one of:

        `doImport`
        `doExport`

    """
    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter=pluginAdapter)

        self._oglObjects:         OglObjects = cast(OglObjects, None)               # The imported Ogl Objects
        self._selectedOglObjects: OglObjects = cast(OglObjects, None)               # The selected Ogl Objects requested by .executeExport()
        self._frameInformation:   FrameInformation = cast(FrameInformation, None)   # The frame information requested by .executeExport()

        #
        # Input pyutplugins that require an active frame or frame(s) should  set this value to `True`
        # Some output pyutplugins may create their own frame or ever their own project and frame.  These should set this value to `False`
        # Plugins should set the value the need in their constructor
        #
        self._requireActiveFrame: bool = True
        #
        # Some Output pyutplugins may offer the option of exporting only selected objects;  Others may just export
        # the entire project or the current frame
        #
        # Plugins should set the value the need in their constructor
        self._requireSelection:   bool = True
        #
        #
        # prefs: PyutPreferences = PyutPreferences()
        # if prefs.pyutIoPluginAutoSelectAll is True:       TODO:  Need plugin preferences
        #     self._autoSelectAll: bool = True
        # else
        self._autoSelectAll: bool = False

    def executeImport(self):
        """
        Called by Pyut to begin the import process.
        """
        # TODO Fix later
        # noinspection PyTypeChecker
        self._pluginAdapter.getFrameInformation(callback=self._executeImport)   # type ignore

    def _executeImport(self, frameInformation: FrameInformation):
        """
        The callback necessary to start the import process;
        Args:
            frameInformation:
        """
        assert self.inputFormat is not None, 'Developer error. We cannot import w/o an import format'
        if self._requireActiveFrame is True:
            if frameInformation.frameActive is False:
                self.displayNoUmlFrame()
                return
        if self.setImportOptions() is True:
            self.read()

    def executeExport(self):
        """
        Called by Pyut to begin the export process.
        """
        if self._autoSelectAll is True:
            self._pluginAdapter.selectAllOglObjects()
        self._pluginAdapter.getFrameInformation(callback=self._executeExport)

    def _executeExport(self, frameInformation: FrameInformation):

        assert self.outputFormat is not None, 'Developer error. We cannot export w/o an output format'

        self._frameInformation = frameInformation

        if frameInformation.frameActive is False:
            self.displayNoUmlFrame()
        else:
            self._selectedOglObjects = frameInformation.selectedOglObjects  # syntactic sugar

            if len(self._selectedOglObjects) == 0 and self._requireSelection is True:
                self.displayNoSelectedOglObjects()
            else:
                if self.setExportOptions() is True:
                    self.write(self._selectedOglObjects)
                    self._pluginAdapter.deselectAllOglObjects()

    @abstractmethod
    def setImportOptions(self) -> bool:
        """
        Prepare for the import.
        Use this method to query the end-user for any additional import options

        Returns:
            if False, the import is cancelled
        """
        pass

    @abstractmethod
    def setExportOptions(self) -> bool:
        """
        Prepare for the export.
        Use this method to query the end-user for any additional export options

        Returns:
            if False, the export is cancelled
        """
        pass

    @abstractmethod
    def read(self) -> bool:
        """
        Read data from a file;  Presumably, the file was specified on the call
        to setImportOptions
        """
        pass

    @abstractmethod
    def write(self, oglObjects: OglObjects):
        """
        Write data to a file;  Presumably, the file was specified on the call
        to setExportOptions

         Args:
            oglObjects:  list of exported objects

        """
        pass




from typing import List

from logging import Logger
from logging import getLogger

from pyutmodel.PyutClass import PyutClass
from ogl.OglClass import OglClass


class ElementTreeData:

    def __init__(self, pyutClass: PyutClass, oglClass: OglClass):

        self.logger: Logger = getLogger(__name__)

        self.pyutClass:         PyutClass  = pyutClass
        self.oglClass:          OglClass   = oglClass
        self._childElementNames: List[str] = []

    def addChild(self, childClassName: str):
        self._childElementNames.append(childClassName)

    @property
    def childElementNames(self) -> List[str]:
        return self._childElementNames

    @childElementNames.setter
    def childElementNames(self, theNewValues: List[str]):
        self._childElementNames = theNewValues

    def __str__(self):
        retStr: str = f'ElementTreeData - ClassName: {self.pyutClass.name} oglClass position: {self.oglClass.GetPosition()}\n'

        for childName in self.childElementNames:
            retStr += f'\t\tchildName: {childName}\n'
        return retStr

    def __repr__(self):
        return self.__str__()
